-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local SpecialSugarVisual = require "GameEx.Components.Visuals.SpecialSugarVisual"
local CircleSelector = require "GameEx.Components.Selectors.CircleSelector"
local CircleImpulser = require "GameEx.Components.Impulsers.CircleImpulser"
local BombRemover = require "GameEx.Components.Removers.BombRemover"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local CombineGenerator = require "GameEx.Components.Generators.CombineGenerator"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"

-- ## Imports ## --




local BombSugar = class("BombSugar", function() return GameObject:create() end)


function BombSugar:ctor()

	local category = self:get_category()
	category:add("special")
	category:add("bomb")

	self:init_state_machine()

	local factory = ObjectFactory:get_instance()

	self:add_component(factory:get("BombRemover"), "remover")
	self:add_component(factory:get("CircleSelector"), "selector")
	self:add_component(factory:get("SpecialSugarVisual"), "visual")
	self:add_component(factory:get("CircleCollider"), "collider")
	self:add_component(factory:get("CircleImpulser"), "impulser")
	self:add_component(factory:get("CombineGenerator"), "generator")

	self:build_filter()

	self:register_listeners_with_config(BombSugar.LISTENER_CONFIG)
	self:register_properties_with_config(BombSugar.PROPERTY_CONFIG)

	self.__c_ignite_action = nil

	self.__n_ignited = false

end

function BombSugar:on_destroy()

	if self.__c_ignite_action then

		self:stopAction(self.__c_ignite_action)

	end

end

function BombSugar:on_reset(c_msg)

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set("key", value)

	end

	self:set("ban_ignite", false)

	self.__n_ignited = false

	self:set("state", "normal")

end

function BombSugar:on_trigger_selector(c_msg)

	self:get_component("selector"):set("select_radius", self:get("normal_radius"))

	local msg = MessagePool:get_free()
	msg.s_name = "select"
	msg.x_param = {ban_message = true, position = c_msg.x_param}

	self:get_component("selector"):send_message(msg)

	c_msg.x_result = msg.x_result

end


function BombSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function BombSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function BombSugar:on_ignite()

	if self:get("ban_ignite") then

		-- self.__c_state_machine:do_event("remove")

	else

		self.__c_state_machine:do_event("ignite")

	end

end

function BombSugar:on_start_remove()

	self.__c_state_machine:do_event("remove")

end

function BombSugar:on_remove_started()
--broadcast to game
	local event = cc.EventCustom:new("bomb_triggered")
	event.param = {}
	cc.Director:getInstance():getEventDispatcher():dispatchEvent(event)

	local msg = MessagePool:get_free()
	msg.s_name = "select"
	msg.x_param = {ban_message = true}

	self:get_component("selector"):send_message(msg)

	local select_result = msg.x_result

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("start_action", "", Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("special_remove_start", {c_list = select_result, t_pos = cc.p(self:getPosition())}, Message.N_TUNNEL))

	local candidate_list = {}

	for _, sugar in ipairs(select_result) do

		if sugar:get_category():has("normal") then

			candidate_list[#candidate_list + 1] = sugar

		end

	end

	self:get_component("generator"):set("candidate_list", candidate_list)

	local msg = MessagePool:get_free()
	msg.s_name = "assign_remove_task"
	msg.x_param = select_result

	self:get_component("remover"):send_message(msg)

	self:send_message(MessagePool:get_free_ex("finish_remove", select_result, Message.N_TUNNEL))

end

function BombSugar:on_ignited()

	local ignite_action = function()

		self.__c_state_machine:do_event("ignite_finish")
		self.__c_ignite_action = nil

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action", "", Message.N_TUNNEL))

	end

	local delay = cc.DelayTime:create(self:get("ignite_delay"));
	local delayed_action = cc.CallFunc:create(ignite_action)

	local sequence = cc.Sequence:create(delay, delayed_action)
	self:runAction(sequence)

	self.__c_ignite_action =  sequence
	self.__n_ignited = true

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("start_action", "", Message.N_TUNNEL))

end

function BombSugar:on_combine(c_msg)

	self:get_component("generator"):set("generate_type", c_msg.x_param)
	self:get_component("generator"):set("enable", true)

	self:get_component("selector"):set("select_radius", self:get("normal_radius"))
	self.__c_state_machine:do_event("remove")	

end

function BombSugar:on_finish_remove(c_msg)

	self:get_component("generator"):set("enable", false)

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("special_remove_finish", {c_list = c_msg.x_param, t_pos = cc.p(self:getPosition())}, Message.N_TUNNEL))

	if self.__n_ignited then

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("bomb_remove_ignite", self, Message.N_TUNNEL))

	else
	
		GameGlobals.Scene:send_message(MessagePool:get_free_ex("bomb_remove", self, Message.N_TUNNEL))

	end

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action", "", Message.N_TUNNEL))

	self:get_component("impulser"):send_message(MessagePool:get_free_ex("apply_impulse", cc.p(self:getPosition()), Message.N_TUNNEL))

end

function BombSugar:on_sugar_remove(c_msg)

	self:get_component("generator"):send_message(MessagePool:get_free_ex("generate", c_msg.x_param))

end

function BombSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function BombSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function BombSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function BombSugar:init_state_machine()

	function on_enter_remove()

		if self.__n_ignited then

			self:get_component("selector"):set("select_radius", self:get("ignite_radius"))

		else

			self:get_component("selector"):set("select_radius", self:get("normal_radius"))

		end

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_enter_ignite()

		self:send_message(MessagePool:get_free_ex("ignited", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "ignite", enter = on_enter_ignite},

	})
	self.__c_state_machine:add_transition({

		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "ignite_finish", from = "ignite", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "unselect", from = "select", to = "normal"},
		{event = "ignite", from = "select", to = "ignite"},
		{event = "ignite", from = "normal", to = "ignite"},

	})
	self.__c_state_machine:do_set("normal")

end

function BombSugar:build_filter()

	local filter = {}

	filter.t_property_filter = {
		type = "and",
		{"state", "=", "normal"},
	}

	filter.t_category_filter = {
		type = "or",
		{"special", "=", true},
		{"normal", "=", true},
		{"fracture", "=", true},
		{"fenced", "=", true},
		{"panda", "=", true},
		{"random", "=", true},
		{"timing_bomb", "=", true},
		{"colorful", "=", true},
	}

	self:get_component("selector"):set("filter", filter)

end

BombSugar.LISTENER_CONFIG = {

	{"ignite", {before = BombSugar.on_ignite}},
	{"select", {before = BombSugar.on_select}},
	{"unselect", {before = BombSugar.on_unselect}},
	{"start_remove", {before = BombSugar.on_start_remove}},
	{"finish_remove", {before = BombSugar.on_finish_remove}},
	{"remove_started", {before = BombSugar.on_remove_started}},
	{"ignited", {before = BombSugar.on_ignited}},
	{"combine", {before = BombSugar.on_combine}},
	{"sugar_remove", {before = BombSugar.on_sugar_remove}},
	{"reset", {before = BombSugar.on_reset}},
	{"destroy", {before = BombSugar.on_destroy}},
	{"trigger_selector", {before = BombSugar.on_trigger_selector}},

}

BombSugar.PROPERTY_CONFIG = {
	
	{"dynamic", true},
	{"color", "bomb"},
	{"normal_radius", 120},
	{"ignite_radius", 240},
	{"impulse_radius", 100},
	{"radius", 60},
	{"state", "normal"},
	{"material", cc.PhysicsMaterial(0.1, 0.0, 0.0)},
	{"ignite_delay", 1.5},
	{"type", "bomb"},
	{"ban_ignite", false},
	{"id", "bomb"},

}


return BombSugar