-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local IceSugarVisual = require "GameEx.Components.Visuals.IceSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"

-- ## Imports ## --




local ChocolateSugar = class("ChocolateSugar", function() return GameObject:create() end)


function ChocolateSugar:ctor()

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("ChocolateSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("chocolate")
	category:add("fracture")
	category:add("rimmed")

	local selector = ObjectFactory:get_instance():get("CircleSelector")
	self:add_component(selector, "selector")

	self:register_listeners_with_config(ChocolateSugar.LISTENER_CONFIG)
	self:register_properties_with_config(ChocolateSugar.PROPERTY_CONFIG)

	self.__n_remove_triggered = false

end

function ChocolateSugar:on_destroy()


end

function ChocolateSugar:on_reset(c_msg)

	self.__n_remove_triggered = false

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")
	self:set("layers", 1)


end

function ChocolateSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function ChocolateSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function ChocolateSugar:on_start_remove(c_msg)

	if not c_msg.n_handled then

		local layer = self:get("layers")

		if layer > 0 then

			layer = layer - 1

			self:send_message(MessagePool:get_free_ex("layer_reduce", "", Message.N_TUNNEL))
			self:set("layers", layer)

			if layer == 0 then

				self.__c_state_machine:do_event("remove")

			end

		end

	end

end

function ChocolateSugar:remove_in_range(remove_list)

	local result = false

	if not self:get_category():has("fenced") then

		local selector = self:get_component("selector")
		selector:set("filter", {})
		selector:set("select_radius", self:get("select_radius"))

		local msg = MessagePool:get_free_ex("select", {ban_message = true})
		selector:send_message(msg)

		local select_result = msg.x_result

		for i_s = 1, #select_result do

			for i_r = 1, #remove_list do

				result = (select_result[i_s] == remove_list[i_r])

				if result then

					return result

				end

			end

		end

	end

	return result

end

function ChocolateSugar:on_link_remove_start(c_msg)
-- print("start_ice_check")
	self.__n_remove_triggered = self:remove_in_range(c_msg.x_param)

end

function ChocolateSugar:on_trigger_indicate(c_msg)

	if self:remove_in_range(c_msg.x_param) then

		self:send_message(MessagePool:get_free_ex("highlight_on", "", Message.N_TUNNEL))

	else

		self:send_message(MessagePool:get_free_ex("highlight_off", "", Message.N_TUNNEL))


	end

end

function ChocolateSugar:on_affect(c_msg)

	local selector = self:get_component("selector")
	selector:set("filter", {

		t_category_filter = {
			type = "and",
			{"normal", "=", true},	
			{"fenced", "=", false},	
			{"random", "=", false},	
			{"timing_bomb", "=", false},	
		},
		t_property_filter = {
			type = "and",
			{"state", "=", "normal"},	
		},

	})
	selector:set("select_radius", self:get("affect_radius"))

	local msg = MessagePool:get_free_ex("select", {ban_message = true})
	selector:send_message(msg)

	local select_result = msg.x_result

	if #select_result > 0 then

		local sugar = select_result[1]
		local sugar_position = cc.p(sugar:getPosition())

		self:send_message(MessagePool:get_free_ex("affect_animation", sugar, Message.N_TUNNEL))

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar_silent", sugar, Message.N_TUNNEL))

		local chocolate = ObjectFactory:get_instance():get("ChocolateSugar")
		chocolate:setPosition(sugar_position)
		GameGlobals.Scene:send_message(MessagePool:get_free_ex("add_sugar", chocolate, Message.N_TUNNEL))

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("chocolate_spread", "", Message.N_TUNNEL))

		c_msg.x_return_value = true

		local function delay()

		end

		RUN_ACTION_WITH_DELAY(1.2, delay)

	else

		c_msg.x_return_value = false

	end

end

function ChocolateSugar:on_link_remove_finish(c_msg)

	if self.__n_remove_triggered then

		self:send_message(MessagePool:get_free_ex("start_remove"))

	end

	self.__n_remove_triggered = false

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("self_remove_finish"))

end

function ChocolateSugar:on_remove_started()

	self:send_message(MessagePool:get_free_ex("finish_remove", "", Message.N_TUNNEL))

end

function ChocolateSugar:on_finish_remove()

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("blast_finale", self, Message.N_TUNNEL))
	
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

	TRIGGER_EVENT("SUGAR_REMOVE", {type = "chocolate"})

end

function ChocolateSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function ChocolateSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)
end

function ChocolateSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function ChocolateSugar:before_set_layers(c_msg)

	if c_msg.x_param.x_new_value < 0 then

		c_msg.x_param.x_new_value = 0

	elseif c_msg.x_param.x_new_value > 3 then

		c_msg.x_param.x_new_value = 3

	end

end

function ChocolateSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},

	})
	self.__c_state_machine:add_transition({

		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end


ChocolateSugar.LISTENER_CONFIG = {

	{"select", {before = ChocolateSugar.on_select}},
	{"unselect", {before = ChocolateSugar.on_unselect}},
	{"start_remove", {after = ChocolateSugar.on_start_remove}},
	{"finish_remove", {before = ChocolateSugar.on_finish_remove}},
	{"remove_started", {before = ChocolateSugar.on_remove_started}},
	{"link_remove_start", {before = ChocolateSugar.on_link_remove_start}},
	{"link_remove_finish", {before = ChocolateSugar.on_link_remove_finish}},
	{"lightning_remove_begin", {before = ChocolateSugar.on_link_remove_start}},
	{"lightning_remove_finish", {before = ChocolateSugar.on_link_remove_finish}},
	{"affect", {before = ChocolateSugar.on_affect}},
	{"reset", {before = ChocolateSugar.on_reset}},
	{"trigger_indicate", {before = ChocolateSugar.on_trigger_indicate}},

}

ChocolateSugar.PROPERTY_CONFIG = {
	
	{"radius", 60},
	{"state", "normal"},
	{"layers", 1},
	{"dynamic", false},
	{"affect_radius", 70},
	{"select_radius", 70},
	{"id", "chocolate"},

}


return ChocolateSugar