-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local CocoSugarVisual = require "GameEx.Components.Visuals.CocoSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"
local ResourceManager = require "GameEx.Foundations.ResourceManager"

-- ## Imports ## --




local CocoSugar = class("CocoSugar", function() return GameObject:create() end)


function CocoSugar:ctor()

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("CocoSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("fracture")
	category:add("coco")
	category:add("rimmed")

	local selector = ObjectFactory:get_instance():get("CircleSelector")
	self:add_component(selector, "selector")

	self:register_listeners_with_config(CocoSugar.LISTENER_CONFIG)
	self:register_properties_with_config(CocoSugar.PROPERTY_CONFIG)

	self.__n_remove_triggered = false

end

function CocoSugar:on_destroy()


end

function CocoSugar:on_reset(c_msg)

	self.__n_remove_triggered = false

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")

end

function CocoSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function CocoSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function CocoSugar:on_start_remove(c_msg)

	local remove_count = 0

	if c_msg.x_param == "normal" then

		remove_count = self.__n_remove_counter

	elseif c_msg.x_param == "nini" then

		remove_count = self:get("accumulate_nini")

	else

		remove_count = self:get("accumulate_special")

	end

	local original_layer = self:get("layers")

	if remove_count > 0 and original_layer > 0 then

		local current_layer = original_layer - remove_count
		current_layer = (current_layer < 0) and 0 or current_layer
		self:set("layers", current_layer)

		self:send_message(MessagePool:get_free_ex("layer_reduce", original_layer, Message.N_TUNNEL))

		if current_layer == 0 then

			self.__c_state_machine:do_event("remove")

		end

	end

	self.__c_state_machine:do_transition("normal")

end

function CocoSugar:remove_in_range(remove_list)

	local counter = 0

	local selector = self:get_component("selector")
	selector:set("position", cc.p(self:getPosition()))
	selector:set("select_radius", self:get("radius") + 50)
	local msg = MessagePool:get_free_ex("select", {ban_message = true})
	selector:send_message(msg)

	local select_result = msg.x_result

	for i_s = 1, #select_result do

		for i_r = 1, #remove_list do

			result = (select_result[i_s] == remove_list[i_r])

			if result then

				counter = counter + self:get("accumulate_normal")

			end

		end

	end

	return counter

end

function CocoSugar:on_link_remove_start(c_msg)

	self.__n_remove_counter = self:remove_in_range(c_msg.x_param)

end

function CocoSugar:on_trigger_indicate(c_msg)

	if self:remove_in_range(c_msg.x_param) > 0 then

		self:send_message(MessagePool:get_free_ex("highlight_on", "", Message.N_TUNNEL))

	else

		self:send_message(MessagePool:get_free_ex("highlight_off", "", Message.N_TUNNEL))

	end

end

function CocoSugar:on_link_remove_finish(c_msg)

	if self.__n_remove_counter > 0 then

		self:send_message(MessagePool:get_free_ex("start_remove", "normal"))

	end

	self.__n_remove_counter = 0

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("self_remove_finish"))

end

function CocoSugar:on_remove_started()



end

function CocoSugar:on_finish_remove()

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action"))

end

function CocoSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function CocoSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function CocoSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function CocoSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "lock"},

	})
	self.__c_state_machine:add_transition({

		{event = "lock", from = "normal", to = "lock"},
		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "remove", from = "lock", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "select", from = "lock", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end

function CocoSugar:on_lock()
	
	self.__c_state_machine:do_event("lock")

end

CocoSugar.LISTENER_CONFIG = {

	{"select", {before = CocoSugar.on_select}},
	{"unselect", {before = CocoSugar.on_unselect}},
	{"start_remove", {after = CocoSugar.on_start_remove}},
	{"finish_remove", {before = CocoSugar.on_finish_remove}},
	{"remove_started", {before = CocoSugar.on_remove_started}},
	{"trigger_indicate", {before = CocoSugar.on_trigger_indicate}},
	{"link_remove_start", {before = CocoSugar.on_link_remove_start}},
	{"link_remove_finish", {before = CocoSugar.on_link_remove_finish}},
	{"lightning_remove_begin", {before = CocoSugar.on_link_remove_start}},
	{"lightning_remove_finish", {before = CocoSugar.on_link_remove_finish}},
	{"reset", {before = CocoSugar.on_reset}},
	{"open", {before = CocoSugar.on_open}},
	{"lock", {before = CocoSugar.on_lock}},

}

CocoSugar.PROPERTY_CONFIG = {
	
	{"state", "normal"},
	{"radius", 124},
	{"id", "coco"},
	{"accumulate_normal", 1},
	{"accumulate_special", 2},
	{"accumulate_nini", 1},
	{"dynamic", false},
	{"layers", 10},

}


return CocoSugar