-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local SpecialSugarVisual = require "GameEx.Components.Visuals.SpecialSugarVisual"
local BoxSelector = require "GameEx.Components.Selectors.BoxSelector"
local CircleImpulser = require "GameEx.Components.Impulsers.CircleImpulser"
local BombRemover = require "GameEx.Components.Removers.BombRemover"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local CombineGenerator = require "GameEx.Components.Generators.CombineGenerator"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"

-- ## Imports ## --




local CrossSugar = class("CrossSugar", function() return GameObject:create() end)


function CrossSugar:ctor()

	local category = self:get_category()
	category:add("special")
	category:add("cross")
	category:add("ten")

	self:init_state_machine()

	local factory = ObjectFactory:get_instance()

	self:add_component(factory:get("BombRemover"), "remover")
	self:add_component(factory:get("BoxSelector"), "selector")
	self:add_component(factory:get("SpecialSugarVisual"), "visual")
	self:add_component(factory:get("CircleCollider"), "collider")
	self:add_component(factory:get("CircleImpulser"), "impulser")
	self:add_component(factory:get("CombineGenerator"), "generator")

	self:build_filter()

	self:register_listeners_with_config(CrossSugar.LISTENER_CONFIG)
	self:register_properties_with_config(CrossSugar.PROPERTY_CONFIG)

	self.__c_ignite_action = nil
	self.__n_ignited = false

end

function CrossSugar:on_destroy()

	if self.__c_ignite_action then

		self:stopAction(self.__c_ignite_action)

	end

end

function CrossSugar:on_reset(c_msg)

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set("key", value)

	end

	self:set("ban_ignite", false)

	self.__n_ignited = false

	self:set("state", "normal")

end

function CrossSugar:on_trigger_selector(c_msg)

	self:set("select_multiplier", 1)

	c_msg.x_result = self:do_select(c_msg.x_param, false)

end

function CrossSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function CrossSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function CrossSugar:on_ignite()

	if self:get("ban_ignite") then

		-- self.__c_state_machine:do_event("remove")

	else

		self.__c_state_machine:do_event("ignite")

	end

end

function CrossSugar:on_start_remove()

	self.__c_state_machine:do_event("remove")

end

function CrossSugar:on_remove_started()

	local select_result = self:do_select(cc.p(self:getPosition()), true)

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("start_action", "", Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("special_remove_start", {c_list = select_result, t_pos = cc.p(self:getPosition())}, Message.N_TUNNEL))

	local candidate_list = {}

	for _, sugar in ipairs(select_result) do

		if sugar:get_category():has("normal") then

			candidate_list[#candidate_list + 1] = sugar

		end

	end

	self:get_component("generator"):set("candidate_list", candidate_list)

	local msg = MessagePool:get_free()
	msg.s_name = "assign_remove_task"
	msg.x_param = select_result

	self:get_component("remover"):send_message(msg)

	self:send_message(MessagePool:get_free_ex("finish_remove", select_result, Message.N_TUNNEL))

end

function CrossSugar:do_select(position, lock_flag)

	local select_result = {}

	local selector = self:get_component("selector")
	selector:set("select_size", cc.size(self:get("select_width"), 1920 * 2))

	local multiplier = self:get("select_multiplier")
	local interval = self:get("select_interval")
	local total_distance = (multiplier - 1) * interval
	local pos = position or cc.p(self:getPosition())
	local current_position_x = pos.x 
	local current_position_y = pos.y

	local start_pos_x = current_position_x - total_distance / 2
	local start_pos_y = current_position_y - total_distance / 2

	for i = 1, multiplier do

		selector:set("position", cc.p(start_pos_x + (i - 1) * interval, start_pos_y))
		selector:set("select_rotation", 0)

		local msg = MessagePool:get_free()
		msg.s_name = "select"
		msg.x_param = {ban_message = true}

		self:get_component("selector"):send_message(msg)

		for i = 1, #msg.x_result do
			
			if lock_flag then msg.x_result[i]:send_message(MessagePool:get_free_ex("lock", "", Message.N_TUNNEL)) end

			select_result[#select_result + 1] = msg.x_result[i]

		end	

	end

	for i = 1, multiplier do

		selector:set("position", cc.p(start_pos_x, start_pos_y + (i - 1) * interval))
		selector:set("select_rotation", 90)

		local msg = MessagePool:get_free()
		msg.s_name = "select"
		msg.x_param = {ban_message = true}

		self:get_component("selector"):send_message(msg)

		for i = 1, #msg.x_result do

			if lock_flag then msg.x_result[i]:send_message(MessagePool:get_free_ex("lock", "", Message.N_TUNNEL)) end
			
			select_result[#select_result + 1] = msg.x_result[i]

		end	

	end

	return select_result	

end

function CrossSugar:on_ignited()

	local ignite_action = function()

		self.__c_state_machine:do_event("ignite_finish")
		self.__c_ignite_action = nil

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action", "", Message.N_TUNNEL))

	end

	local delay = cc.DelayTime:create(self:get("ignite_delay"));
	local delayed_action = cc.CallFunc:create(ignite_action)


	local sequence = cc.Sequence:create(delay, delayed_action)
	self:runAction(sequence)

	self.__c_ignite_action =  sequence
	self.__n_ignited = true

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("start_action", "", Message.N_TUNNEL))

end

function CrossSugar:on_combine(c_msg)

	self:get_component("generator"):set("generate_type", c_msg.x_param)
	self:get_component("generator"):set("enable", true)

	self:set("select_multiplier", 1)
	self.__c_state_machine:do_event("remove")	

end

function CrossSugar:on_finish_remove(c_msg)

	self:get_component("generator"):set("enable", false)

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("special_remove_finish", {c_list = c_msg.x_param, t_pos = cc.p(self:getPosition())}, Message.N_TUNNEL))

	if self.__n_ignited then

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("cross_remove_ignite", self, Message.N_TUNNEL))

	else
	
		GameGlobals.Scene:send_message(MessagePool:get_free_ex("cross_remove", self, Message.N_TUNNEL))

	end

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action", "", Message.N_TUNNEL))

end

function CrossSugar:on_sugar_remove(c_msg)

	self:get_component("generator"):send_message(MessagePool:get_free_ex("generate", c_msg.x_param))

end

function CrossSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function CrossSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function CrossSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function CrossSugar:init_state_machine()

	function on_enter_remove()

		if self.__n_ignited then

			self:set("select_multiplier", 2)

		else

			self:set("select_multiplier", 1)

		end

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_enter_ignite()

		self:send_message(MessagePool:get_free_ex("ignited", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "ignite", enter = on_enter_ignite},

	})
	self.__c_state_machine:add_transition({

		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "ignite_finish", from = "ignite", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "unselect", from = "select", to = "normal"},
		{event = "ignite", from = "select", to = "ignite"},
		{event = "ignite", from = "normal", to = "ignite"},

	})
	self.__c_state_machine:do_set("normal")

end

function CrossSugar:build_filter()

	local filter = {}

	filter.t_property_filter = {
		type = "and",
		{"state", "=", "normal"},
	}

	filter.t_category_filter = {
		type = "or",
		{"special", "=", true},
		{"normal", "=", true},
		{"fracture", "=", true},
		{"fenced", "=", true},
		{"panda", "=", true},
		{"random", "=", true},
		{"timing_bomb", "=", true},
		{"colorful", "=", true},
	}

	self:get_component("selector"):set("filter", filter)

end

CrossSugar.LISTENER_CONFIG = {

	{"ignite", {before = CrossSugar.on_ignite}},
	{"select", {before = CrossSugar.on_select}},
	{"unselect", {before = CrossSugar.on_unselect}},
	{"start_remove", {before = CrossSugar.on_start_remove}},
	{"finish_remove", {before = CrossSugar.on_finish_remove}},
	{"remove_started", {before = CrossSugar.on_remove_started}},
	{"ignited", {before = CrossSugar.on_ignited}},
	{"combine", {before = CrossSugar.on_combine}},
	{"sugar_remove", {before = CrossSugar.on_sugar_remove}},
	{"reset", {before = CrossSugar.on_reset}},
	{"destroy", {before = CrossSugar.on_destroy}},
	{"trigger_selector", {before = CrossSugar.on_trigger_selector}},

}

CrossSugar.PROPERTY_CONFIG = {
	
	{"dynamic", true},
	{"color", "ten"},
	{"select_width", 15},
	{"select_multiplier", 1},
	{"select_interval", 240},
	{"radius", 60},
	{"state", "normal"},
	{"material", cc.PhysicsMaterial(0.1, 0.0, 0.0)},
	{"ignite_delay", 1.5},
	{"type", "ten"},
	{"ban_ignite", false},
	{"id", "cross"}

}


return CrossSugar