-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local IceSugarVisual = require "GameEx.Components.Visuals.IceSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"

-- ## Imports ## --




local FanSugar = class("FanSugar", function() return GameObject:create() end)


function FanSugar:ctor()

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("FanSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("fan")
	category:add("rimmed")

	local selector = ObjectFactory:get_instance():get("CircleSelector")
	self:add_component(selector, "selector")

	self:register_listeners_with_config(FanSugar.LISTENER_CONFIG)
	self:register_properties_with_config(FanSugar.PROPERTY_CONFIG)

	self.__n_remove_color = false

	self.__colors = {}

end

function FanSugar:on_destroy()


end

function FanSugar:on_reset(c_msg)

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")

	self.__n_remove_color = nil
	self.__colors = {}

end

function FanSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function FanSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function FanSugar:on_start_remove()

	local color_duplicated = false

	for i = 1, #self.__colors do

		if self.__colors[i] == self.__n_remove_color then

			color_duplicated = true

		end

	end

	if not color_duplicated then

		self.__colors[#self.__colors + 1] = self.__n_remove_color

		self:send_message(MessagePool:get_free_ex("fill_color", self.__n_remove_color, Message.N_TUNNEL))

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("flower_fill", "", Message.N_TUNNEL))

		if #self.__colors == self:get("volume") then

			self.__c_state_machine:do_event("remove")

		end

	end

end

function FanSugar:on_link_remove_start(c_msg)

	local remove_list = c_msg.x_param
	local selector = self:get_component("selector")
	selector:set("filter", {})
	local msg = MessagePool:get_free_ex("select", {ban_message = true})
	selector:send_message(msg)

	local select_result = msg.x_result

	for i_s = 1, #select_result do

		for i_r = 1, #remove_list do

			local in_range = (select_result[i_s] == remove_list[i_r])

			if in_range then

				self.__n_remove_color = remove_list[i_r]:get("color")

				return

			end

		end

	end

end

function FanSugar:on_link_remove_finish(c_msg)

	if self.__n_remove_color then

		self:on_start_remove()

	end

	self.__n_remove_color = nil

end

function FanSugar:on_remove_started()

	self:send_message(MessagePool:get_free_ex("finish_remove", "", Message.N_TUNNEL))

end

function FanSugar:on_finish_remove()

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

end

function FanSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function FanSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function FanSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end


function FanSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},

	})
	self.__c_state_machine:add_transition({

		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end


FanSugar.LISTENER_CONFIG = {

	{"select", {before = FanSugar.on_select}},
	{"unselect", {before = FanSugar.on_unselect}},
	{"start_remove", {before = FanSugar.on_start_remove}},
	{"finish_remove", {before = FanSugar.on_finish_remove}},
	{"remove_started", {before = FanSugar.on_remove_started}},
	{"link_remove_start", {before = FanSugar.on_link_remove_start}},
	{"link_remove_finish", {before = FanSugar.on_link_remove_finish}},
	{"lightning_remove_begin", {before = FanSugar.on_link_remove_start}},
	{"lightning_remove_finish", {before = FanSugar.on_link_remove_finish}},
	{"reset", {before = FanSugar.on_reset}}

}

FanSugar.PROPERTY_CONFIG = {
	
	{"volume", 0},
	{"radius", 90},
	{"state", "normal"},
	{"dynamic", false},
	{"width", 90},
	{"height", 90},
	{"id", "fan"},

}


return FanSugar