-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local GiftSugarVisual = require "GameEx.Components.Visuals.GiftSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"
local ResourceManager = require "GameEx.Foundations.ResourceManager"

-- ## Imports ## --




local GiftSugar = class("GiftSugar", function() return GameObject:create() end)


function GiftSugar:ctor()

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("GiftSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("fracture")
	category:add("gift")
	category:add("rimmed")

	local selector = ObjectFactory:get_instance():get("CircleSelector")
	self:add_component(selector, "selector")

	self:register_listeners_with_config(GiftSugar.LISTENER_CONFIG)
	self:register_properties_with_config(GiftSugar.PROPERTY_CONFIG)

	self.__n_remove_triggered = false
	self.__n_accumulator = 0

end

function GiftSugar:on_destroy()


end

function GiftSugar:on_reset(c_msg)

	self.__n_remove_triggered = false

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")

	self.__n_accumulator = 0

end

function GiftSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function GiftSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function GiftSugar:on_start_remove(c_msg)

	if c_msg.x_param == "normal" then

		self.__n_accumulator = self.__n_accumulator + self.__n_remove_counter

	elseif c_msg.x_param == "nini" then

		self.__n_accumulator = self.__n_accumulator + self:get("accumulate_nini")

	else

		self.__n_accumulator = self.__n_accumulator + self:get("accumulate_special")

	end

	self:send_message(MessagePool:get_free_ex("accumulate", self.__n_accumulator, Message.N_TUNNEL))

	self.__c_state_machine:do_transition("normal")

end

function GiftSugar:remove_in_range(remove_list)

	local counter = 0

	local selector = self:get_component("selector")
	selector:set("position", cc.p(self:getPosition()))
	selector:set("select_radius", self:get("radius") + 50)
	local msg = MessagePool:get_free_ex("select", {ban_message = true})
	selector:send_message(msg)

	local select_result = msg.x_result

	for i_s = 1, #select_result do

		for i_r = 1, #remove_list do

			result = (select_result[i_s] == remove_list[i_r])

			if result then

				counter = counter + self:get("accumulate_normal")

			end

		end

	end

	return counter

end

function GiftSugar:on_link_remove_start(c_msg)

	self.__n_remove_counter = self:remove_in_range(c_msg.x_param)

end

function GiftSugar:on_trigger_indicate(c_msg)

	if self:remove_in_range(c_msg.x_param) > 0 then

		self:send_message(MessagePool:get_free_ex("highlight_on", "", Message.N_TUNNEL))

	else

		self:send_message(MessagePool:get_free_ex("highlight_off", "", Message.N_TUNNEL))

	end

end

function GiftSugar:on_link_remove_finish(c_msg)

	if self.__n_remove_counter > 0 then

		self:send_message(MessagePool:get_free_ex("start_remove", "normal"))

	end

	self.__n_remove_counter = 0

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("self_remove_finish"))

end

function GiftSugar:on_remove_started()

    local t_filter = {}

    t_filter.t_category_filter = {
        type = "and",
        {"normal", "=", true},
        {"fenced", "=", false},
        {"add_move", "=", false},    
    }
    t_filter.t_property_filter = {
        type = "and",
        {"state", "=", "normal"},
    }

    local param = {

        t_category_filter = t_filter.t_category_filter,
        t_property_filter = t_filter.t_property_filter,

    }

    local msg = MessagePool:get_free()
    msg.c_sender = self
    msg.s_name = "query_filter"
    msg.x_param = param

    GameGlobals.Scene:send_message(msg)

    if #msg.x_result > 0 then

        local counter = self:get("bomb_number") + self:get("cross_number") + self:get("add_move_number")
        local count = 0

        local type_map = {}

        for i = 1, self:get("bomb_number") do

        	type_map[i] = "BombSugar"

        end

        for i = #type_map + 1, #type_map + self:get("cross_number") do

        	type_map[i] = "CrossSugar"

        end

        for i = #type_map + 1, #type_map + self:get("add_move_number") do

        	type_map[i] = "AddMoveSugar"

        end

        if #msg.x_result < counter then counter = #msg.x_result end

        for i = 1, counter do 

            local target = msg.x_result[i]

            target:send_message(MessagePool:get_free_ex("lock"))

            local visible_size = cc.Director:getInstance():getVisibleSize()

            local homing_entry = 0
            local homing_particle = ResourceManager:get_instance():get_resource_by_key("skill_flying_particle")
            homing_particle:setDuration(-1)
            homing_particle:setPosition(cc.p(self:getPosition()))
            homing_particle:setLocalZOrder(11)
            homing_particle:setAutoRemoveOnFinish(false)
            homing_particle:setSpeed(100)
            -- homing_particle:setPosition(cc.p(msg.x_result[i]:getPosition()))
            GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", homing_particle))

            local gen_type = type_map[i]

            local homing_performer = function()

                local target_position = target:getParent():convertToWorldSpace(cc.p(target:getPosition()))
                local particle_position = cc.p(homing_particle:getPosition())

                local distance = UtilFunction:caculate_distance(target_position, particle_position)

                if distance < self:get("homing_speed") then

	                local particle = ResourceManager:get_instance():get_resource_by_key("bonus_remove_particle")
	                particle:setLocalZOrder(99)
	                particle:setPosition(target_position)
	                particle:setAutoRemoveOnFinish(true)

	                GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", particle))

                    local sugar = target

                    local position = cc.p(sugar:getPosition())

                    GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar_silent", sugar, Message.N_TUNNEL))

                    GameGlobals.Scene:send_message(MessagePool:get_free_ex("bonus_replace_gift", target, Message.N_TUNNEL))

                    local line = ObjectFactory:get_instance():get(gen_type)
                    -- line:set("color", sugar:get("color"))
                    line:setPosition(position)
                    GameGlobals.Scene:send_message(MessagePool:get_free_ex("add_sugar", line, Message.N_TUNNEL))

                    homing_particle:setDuration(0.1)
                    homing_particle:setAutoRemoveOnFinish(true)

                    cc.Director:getInstance():getScheduler():unscheduleScriptEntry(homing_entry)

                    count = count + 1

                    if count == counter then

                        GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action", task, Message.N_TUNNEL))
                        self:send_message(MessagePool:get_free_ex("finish_remove", "", Message.N_TUNNEL))

                    end

                else

                    local speed = self:get("homing_speed")
                    local direction = UtilFunction:caculate_direction(target_position, particle_position)
                    local angle = UtilFunction:caculate_angle(direction)
                    local offset = cc.p(speed * math.cos(math.rad(-angle)), speed * math.sin(math.rad(-angle)))
                    particle_position.x = particle_position.x + offset.x
                    particle_position.y = particle_position.y + offset.y

                    homing_particle:setAngle(-angle)
                    homing_particle:setPosition(particle_position)

                end

            end

            homing_entry = cc.Director:getInstance():getScheduler():scheduleScriptFunc(homing_performer, 0.016, false)

        end

        GameGlobals.Scene:send_message(MessagePool:get_free_ex("start_action"))

    end

end

function GiftSugar:on_finish_remove()

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

end

function GiftSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function GiftSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function GiftSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function GiftSugar:on_open(c_msg)

	if self.__n_accumulator >= 100 then
-- ERROR_ASSERT(false, "string")
		self.__c_state_machine:do_event("remove")

	end

end

function GiftSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "lock"},

	})
	self.__c_state_machine:add_transition({

		{event = "lock", from = "normal", to = "lock"},
		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "remove", from = "lock", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "select", from = "lock", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end

function GiftSugar:on_lock()
	
	self.__c_state_machine:do_event("lock")

end

GiftSugar.LISTENER_CONFIG = {

	{"select", {before = GiftSugar.on_select}},
	{"unselect", {before = GiftSugar.on_unselect}},
	{"start_remove", {after = GiftSugar.on_start_remove}},
	{"finish_remove", {before = GiftSugar.on_finish_remove}},
	{"remove_started", {before = GiftSugar.on_remove_started}},
	{"trigger_indicate", {before = GiftSugar.on_trigger_indicate}},
	{"link_remove_start", {before = GiftSugar.on_link_remove_start}},
	{"link_remove_finish", {before = GiftSugar.on_link_remove_finish}},
	{"lightning_remove_begin", {before = GiftSugar.on_link_remove_start}},
	{"lightning_remove_finish", {before = GiftSugar.on_link_remove_finish}},
	{"reset", {before = GiftSugar.on_reset}},
	{"open", {before = GiftSugar.on_open}},
	{"lock", {before = GiftSugar.on_lock}},

}

GiftSugar.PROPERTY_CONFIG = {
	
	{"state", "normal"},
	{"radius", 124},
	{"id", "gift"},
	{"accumulate_normal", 10},
	{"accumulate_special", 15},
	{"accumulate_nini", 30},
	{"bomb_number", 3},
	{"cross_number", 1},
	{"add_move_number", 1},
	{"move_sugar", 1},
	{"dynamic", false},
	{"homing_speed", 25},

}


return GiftSugar