-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local SpecialSugarVisual = require "GameEx.Components.Visuals.SpecialSugarVisual"
local ColorSelector = require "GameEx.Components.Selectors.ColorSelector"
local TotalSelector = require "GameEx.Components.Selectors.TotalSelector"
local CircleImpulser = require "GameEx.Components.Impulsers.CircleImpulser"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local CombineGenerator = require "GameEx.Components.Generators.CombineGenerator"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"

-- ## Imports ## --




local LightningSugar = class("LightningSugar", function() return GameObject:create() end)


function LightningSugar:ctor()

	local category = self:get_category()
	category:add("special")
	category:add("lightning")

	self:init_state_machine()

	local factory = ObjectFactory:get_instance()

	self.__c_color_selector = factory:get("ColorSelector")
	self.__c_color_selector:retain()

	self.__c_total_selector = factory:get("TotalSelector")
	self.__c_total_selector:retain()

	self.__c_color_remover = factory:get("LightningRemover")
	self.__c_color_remover:retain()

	self.__c_total_remover = factory:get("TotalRemover")
	self.__c_total_remover:retain()

	self:add_component(factory:get("SpecialSugarVisual"), "visual")
	self:add_component(factory:get("CircleCollider"), "collider")
	self:add_component(factory:get("CircleImpulser"), "impulser")
	self:add_component(factory:get("CombineGenerator"), "generator")

	self:register_listeners_with_config(LightningSugar.LISTENER_CONFIG)
	self:register_properties_with_config(LightningSugar.PROPERTY_CONFIG)

	self:change_to_color()

	self.__c_ignite_action = nil

end

function LightningSugar:on_destroy()

	if self.__c_ignite_action then

		self:stopAction(self.__c_ignite_action)

	end

	self.__c_color_selector:release()
	self.__c_total_selector:release()
	self.__c_color_remover:release()
	self.__c_total_remover:release()

end

function LightningSugar:on_reset(c_msg)

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set("key", value)

	end

	self:set("state", "normal")

end

function LightningSugar:on_trigger_selector(c_msg)

	local msg = MessagePool:get_free()
	msg.s_name = "select"
	msg.x_param = {}
	msg.x_param.position = cc.p(self:getPosition())
	msg.x_param.ban_message = true

	self:get_component("selector"):send_message(msg)

	c_msg.x_result = msg.x_result

end

function LightningSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function LightningSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function LightningSugar:on_ignite()

	self:change_to_color()

	self.__c_state_machine:do_event("ignite")

end

function LightningSugar:on_start_remove()

	self:change_to_color()

	self.__c_state_machine:do_event("remove")

end

function LightningSugar:on_remove_started()

	local function perform_remove()

		local msg = MessagePool:get_free()
		msg.s_name = "select"
		msg.x_param = {}
		msg.x_param.position = cc.p(self:getPosition())
		msg.x_param.ban_message = true

		self:get_component("selector"):send_message(msg)

		local select_result = msg.x_result

		for i = 1, #select_result do

			select_result[i]:send_message(MessagePool:get_free_ex("lock", "", Message.N_TUNNEL))

		end

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("lightning_remove_begin", select_result, Message.N_TUNNEL))
		-- GameGlobals.Scene:send_message_to_sugars(MessagePool:get_free_ex("lightning_remove_begin", select_result, Message.N_TUNNEL))

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("special_remove_start", {c_list = select_result, t_pos = cc.p(self:getPosition())}, Message.N_TUNNEL))

		local candidate_list = {}

		for _, sugar in ipairs(select_result) do

			if sugar:get_category():has("normal") then

				candidate_list[#candidate_list + 1] = sugar

			end

		end

		self:get_component("generator"):set("candidate_list", candidate_list)

		local msg = MessagePool:get_free()
		msg.s_name = "assign_remove_task"
		msg.x_param = select_result

		self:get_component("remover"):send_message(msg)

	end

	self:runAction(cc.Sequence:create(cc.DelayTime:create(self:get("start_delay")), cc.CallFunc:create(perform_remove)));

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("start_action", task, Message.N_TUNNEL))

end

function LightningSugar:on_ignited()

	local ignite_action = function()

		self.__c_state_machine:do_event("ignite_finish")
		self.__c_ignite_action = nil

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action", "", Message.N_TUNNEL))

	end

	local delay = cc.DelayTime:create(self:get("ignite_delay"));
	local delayed_action = cc.CallFunc:create(ignite_action)


	local sequence = cc.Sequence:create(delay, delayed_action)
	self:runAction(sequence)

	self.__c_ignite_action =  sequence

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("start_action", "", Message.N_TUNNEL))

end

function LightningSugar:on_combine(c_msg)

	if c_msg.x_param == "lightning" then

		self:change_to_total()
		self.__c_state_machine:do_event("remove")

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("total_remove_start", "", Message.N_TUNNEL))

	else

		self:get_component("generator"):set("generate_type", c_msg.x_param)
		self:get_component("generator"):set("enable", true)

		self:change_to_color()

		self.__c_state_machine:do_event("remove")

	end	

end

function LightningSugar:change_to_total()

	self:remove_component("selector")
	self:add_component(self.__c_total_selector, "selector")

	self:remove_component("remover")
	self:add_component(self.__c_total_remover, "remover")

	self:set("total", true)	
	self:set("start_delay", 40 * 0.033)

	self:setLocalZOrder(99)

end

function LightningSugar:change_to_color()

	self:remove_component("selector")
	self:add_component(self.__c_color_selector, "selector")

	self:remove_component("remover")
	self:add_component(self.__c_color_remover, "remover")

	self:set("total", false)
	self:set("start_delay", 3 * 0.033)		

end

function LightningSugar:on_finish_remove(c_msg)

	self:get_component("generator"):set("enable", false)

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("special_remove_finish", {c_list = c_msg.x_param, t_pos = cc.p(self:getPosition())}, Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("finish_action", "", Message.N_TUNNEL))

	if self:get("total") then

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("lightning_remove_ignite", self, Message.N_TUNNEL))

	else
	
		GameGlobals.Scene:send_message(MessagePool:get_free_ex("lightning_remove", self, Message.N_TUNNEL))

	end

end

function LightningSugar:on_sugar_remove(c_msg)

	self:get_component("generator"):send_message(MessagePool:get_free_ex("generate", c_msg.x_param))

end

function LightningSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function LightningSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function LightningSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function LightningSugar:after_set_select_color(c_msg)

	self.__c_color_selector:set("select_color", c_msg.x_param.x_new_value)

end

function LightningSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_enter_ignite()

		self:send_message(MessagePool:get_free_ex("ignited", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "ignite", enter = on_enter_ignite},

	})
	self.__c_state_machine:add_transition({

		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "ignite_finish", from = "ignite", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "unselect", from = "select", to = "normal"},
		{event = "ignite", from = "select", to = "ignite"},
		{event = "ignite", from = "normal", to = "ignite"},

	})
	self.__c_state_machine:do_set("normal")

end

function LightningSugar:on_special_triggered(c_msg)

	self:change_to_color()

end

function LightningSugar:on_special_link_triggered(c_msg)
print("lightning special link triggered")
	local task_list = c_msg.x_param

	local total_flag = true

	for i = 1, #task_list do

		total_flag = total_flag and task_list[i]:get_category():has("lightning")

	end

	if total_flag then

		self:change_to_total()

	else

		self:change_to_color()

	end

end

LightningSugar.LISTENER_CONFIG = {

	{"ignite", {before = LightningSugar.on_ignite}},
	{"select", {before = LightningSugar.on_select}},
	{"unselect", {before = LightningSugar.on_unselect}},
	{"start_remove", {before = LightningSugar.on_start_remove}},
	{"finish_remove", {before = LightningSugar.on_finish_remove}},
	{"remove_started", {before = LightningSugar.on_remove_started}},
	{"ignited", {before = LightningSugar.on_ignited}},
	{"combine", {before = LightningSugar.on_combine}},
	{"sugar_remove", {before = LightningSugar.on_sugar_remove}},
	{"reset", {before = LightningSugar.on_reset}},
	{"destroy", {before = LightningSugar.on_destroy}},
	{"trigger_selector", {before = LightningSugar.on_trigger_selector}},
	{"special_link_triggered", {before = LightningSugar.on_special_link_triggered}},
	{"special_triggered", {before = LightningSugar.on_special_triggered}},

}

LightningSugar.PROPERTY_CONFIG = {
	
	{"dynamic", true},
	{"color", "lightning"},
	{"impulse_radius", 100},
	{"radius", 60},
	{"state", "normal"},
	{"material", cc.PhysicsMaterial(0.1, 0.0, 0.0)},
	{"ignite_delay", 1.5},
	{"type", "lightning"},
	{"start_delay", 0},
	{"select_color", "red"},
	{"id", "lightning"},
}


return LightningSugar