-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local NormalSugarVisual = require "GameEx.Components.Visuals.NormalSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"
local ProbGetter = require "GameEx.Foundations.Utils.ProbGetter"
local RandomGenerator = require "GameEx.Foundations.Utils.RandomGenerator"

-- ## Imports ## --




local NormalSugar = class("NormalSugar", function() return GameObject:create() end)

function NormalSugar:ctor()

	self.__c_prob_getter = ProbGetter:create(GameGlobals.RandomGenerator)

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("NormalSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("normal")

	self.__n_random_flag = false

	self:register_listeners_with_config(NormalSugar.LISTENER_CONFIG)
	self:register_properties_with_config(NormalSugar.PROPERTY_CONFIG)

end

function NormalSugar:on_destroy(c_msg)


end

function NormalSugar:on_enter()

	if self.__n_random_flag then

		local new_color = self.__c_prob_getter:get()

		self:set("color", new_color)
		self:get_category():add(new_color)

	end

	self.__n_random_flag = false

end

function NormalSugar:on_reset(c_msg)

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")

end

function NormalSugar:on_lock()
	
	self.__c_state_machine:do_event("lock")

end

function NormalSugar:on_select()
	print("select")
	self.__c_state_machine:do_event("select")

end

function NormalSugar:on_unselect()
	print("unselect")
	self.__c_state_machine:do_event("unselect")

end

function NormalSugar:on_start_remove(c_msg)

	if not c_msg.n_handled then

		self.__c_state_machine:do_event("remove")

	end

end

function NormalSugar:on_remove_started(c_msg)

	self:send_message(MessagePool:get_free_ex("finish_remove", "", Message.N_TUNNEL))

end

function NormalSugar:on_finish_remove()
-- 
	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

end

function NormalSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function NormalSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function NormalSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "lock"},

	})
	self.__c_state_machine:add_transition({

		{event = "lock", from = "normal", to = "lock"},
		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "remove", from = "lock", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "select", from = "lock", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end

function NormalSugar:after_set_fenced(c_msg)

	if c_msg.x_param.x_new_value == true then

		if self:get_component("fence") == GameObject.C_EMPTY_COMPONENT then

			local fence = ObjectFactory:get_instance():get("FenceSugar")
			fence:send_message(MessagePool:get_free_ex("append_to", self, Message.N_TUNNEL))

		end

	else

		if self:get_component("fence") then

			self:remove_component("fence")

		end

	end

end

function NormalSugar:before_set_color(c_msg)

	local last_color = self:get("color")
	self:get_category():remove(last_color)

	local new_color = c_msg.x_param.x_new_value

	if new_color == "random" then

		self.__n_random_flag = true

	else

		self.__n_random_flag = false
		self:get_category():add(new_color)

	end

end

function NormalSugar:after_set_prob_red(c_msg)

	self:change_prob("red", c_msg.x_param.x_new_value)

end

function NormalSugar:after_set_prob_green(c_msg)

	self:change_prob("green", c_msg.x_param.x_new_value)

end

function NormalSugar:after_set_prob_yellow(c_msg)

	self:change_prob("yellow", c_msg.x_param.x_new_value)

end

function NormalSugar:after_set_prob_pink(c_msg)

	self:change_prob("pink", c_msg.x_param.x_new_value)

end

function NormalSugar:after_set_prob_blue(c_msg)

	self:change_prob("blue", c_msg.x_param.x_new_value)

end

function NormalSugar:after_set_prob_white(c_msg)

	self:change_prob("white", c_msg.x_param.x_new_value)

end

function NormalSugar:change_prob(s_key, n_prob)

	if self.__c_prob_getter:has(s_key) then

		self.__c_prob_getter:modify(s_key, n_prob)

	else

		self.__c_prob_getter:add(s_key, s_key, n_prob)

	end	

end

function NormalSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function NormalSugar:after_set_tagged(c_msg)

	if c_msg.x_param.x_new_value == true then

		self:get_category():add("tagged")

	else

		self:get_category():remove("tagged")

	end

end


NormalSugar.LISTENER_CONFIG = {

	{"lock", {before = NormalSugar.on_lock}},
	{"select", {before = NormalSugar.on_select}},
	{"unselect", {before = NormalSugar.on_unselect}},
	{"start_remove", {after = NormalSugar.on_start_remove}},
	{"finish_remove", {before = NormalSugar.on_finish_remove}},
	{"remove_started", {before = NormalSugar.on_remove_started}},
	{"target", {before = NormalSugar.on_target}},
	{"reset", {before = NormalSugar.on_reset}},
	{"enter", {before = NormalSugar.on_enter}}

}

NormalSugar.PROPERTY_CONFIG = {
	
	{"dynamic", true},
	{"radius", 60},
	{"state", "normal"},
	{"fenced", false},
	{"prob_red", 0},
	{"prob_green", 0},
	{"prob_blue", 0},
	{"prob_yellow", 0},
	{"prob_pink", 10},
	{"prob_white", 0},
	{"color", "blue"},
	{"material", cc.PhysicsMaterial(0.1, 0.4, 0.1)},
	{"id", "-"},
	{"tagged", false},

}


return NormalSugar