-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local TimingBombSugarVisual = require "GameEx.Components.Visuals.TimingBombSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"
local ProbGetter = require "GameEx.Foundations.Utils.ProbGetter"
local RandomGenerator = require "GameEx.Foundations.Utils.RandomGenerator"
local ConfigManager = require "GameEx.Foundations.ConfigManager"

-- ## Imports ## --




local RandomSugar = class("RandomSugar", function() return GameObject:create() end)



function RandomSugar:ctor()

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("RandomSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("random")

	self.__n_random_flag = false

	self.__c_random_generator = RandomGenerator:create()
	self.__c_generate_prob_getter = ProbGetter:create(GameGlobals.RandomGenerator)
	self.__c_color_prob_getter = ProbGetter:create(GameGlobals.RandomGenerator)

	self:register_listeners_with_config(RandomSugar.LISTENER_CONFIG)
	self:register_properties_with_config(RandomSugar.PROPERTY_CONFIG)

end

function RandomSugar:on_destroy()



end

function RandomSugar:on_enter()

	if self.__n_random_flag then

		self:set("color", self.__c_color_prob_getter:get())

	end

	self.__n_random_flag = false

end

function RandomSugar:on_reset(c_msg)

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")

end

function RandomSugar:on_lock()
	
	self.__c_state_machine:do_event("lock")

end

function RandomSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function RandomSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function RandomSugar:on_start_remove(c_msg)

	if not c_msg.n_handled then

		self.__c_state_machine:do_event("remove")

	end

end

function RandomSugar:on_remove_started(c_msg)

	

end

function RandomSugar:on_finish_remove()

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

end

function RandomSugar:generate()

	local ref = self.__c_generate_prob_getter:get()

	if ref then

		local config = ConfigManager:get_instance():get_config(ref.type)
		local properties = {} 

		for _, tile in ipairs(config.tiles) do

			if tile.id == ref.tile_index then

				properties = tile.properties

			end

		end 

		local sugar = ObjectFactory:get_instance():get(ref.type)

		for key, value in pairs(properties) do

			sugar:set(key, value)

		end

		sugar:setPosition(self:getPosition())

		GameGlobals.Scene:send_message(MessagePool:get_free_ex("add_sugar", sugar, Message.N_TUNNEL))

		if sugar:get_category():has("chocolate") then

			GameGlobals.Scene:send_message(MessagePool:get_free_ex("chocolate_affect", "chocolate", Message.N_TUNNEL))

		end

	end

end

function RandomSugar:after_set_ref(c_msg)

	self.__c_generate_prob_getter:clear()

	local ref_config = c_msg.x_param.x_new_value

	for index, ref in ipairs(ref_config) do 

		self.__c_generate_prob_getter:add(index, ref, ref.prob)

	end

end

function RandomSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function RandomSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function RandomSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function RandomSugar:before_set_color(c_msg)

	if c_msg.x_param.x_new_value == "random" then

		self.__n_random_flag = true

	else

		self.__n_random_flag = false

	end

end

function RandomSugar:after_set_prob_red(c_msg)

	self:change_prob("red", c_msg.x_param.x_new_value)

end

function RandomSugar:after_set_prob_green(c_msg)

	self:change_prob("green", c_msg.x_param.x_new_value)

end

function RandomSugar:after_set_prob_yellow(c_msg)

	self:change_prob("yellow", c_msg.x_param.x_new_value)

end

function RandomSugar:after_set_prob_pink(c_msg)

	self:change_prob("pink", c_msg.x_param.x_new_value)

end

function RandomSugar:after_set_prob_blue(c_msg)

	self:change_prob("blue", c_msg.x_param.x_new_value)

end

function RandomSugar:after_set_prob_white(c_msg)

	self:change_prob("white", c_msg.x_param.x_new_value)

end

function RandomSugar:change_prob(s_key, n_prob)

	if self.__c_color_prob_getter:has(s_key) then

		self.__c_color_prob_getter:modify(s_key, n_prob)

	else

		self.__c_color_prob_getter:add(s_key, s_key, n_prob)

	end	

end

function RandomSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "lock"},

	})
	self.__c_state_machine:add_transition({

		{event = "lock", from = "normal", to = "lock"},
		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "remove", from = "lock", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "select", from = "lock", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end


RandomSugar.LISTENER_CONFIG = {

	{"lock", {before = RandomSugar.on_lock}},
	{"select", {before = RandomSugar.on_select}},
	{"unselect", {before = RandomSugar.on_unselect}},
	{"start_remove", {after = RandomSugar.on_start_remove}},
	{"finish_remove", {before = RandomSugar.on_finish_remove}},
	{"remove_finished", {before = RandomSugar.generate}},
	{"remove_started", {before = RandomSugar.on_remove_started}},
	{"target", {before = RandomSugar.on_target}},
	{"reset", {before = RandomSugar.on_reset}},
	{"enter", {before = RandomSugar.on_enter}},

}

RandomSugar.PROPERTY_CONFIG = {
	
	{"dynamic", true},
	{"prob_red", 0},
	{"prob_green", 0},
	{"prob_blue", 10},
	{"prob_yellow", 10},
	{"prob_pink", 0},
	{"prob_white", 0},
	{"color", "random"},
	{"radius", 60},
	{"state", "normal"},
	{"ref", {}},
	{"id", "random"}

}


return RandomSugar