-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local IceSugarVisual = require "GameEx.Components.Visuals.IceSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"

-- ## Imports ## --




local RewardBagSugar = class("RewardBagSugar", function() return GameObject:create() end)


function RewardBagSugar:ctor()

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("RewardBagSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("fracture")
	category:add("reward_bag")
	category:add("rimmed")

	local selector = ObjectFactory:get_instance():get("CircleSelector")
	self:add_component(selector, "selector")

	self:register_listeners_with_config(RewardBagSugar.LISTENER_CONFIG)
	self:register_properties_with_config(RewardBagSugar.PROPERTY_CONFIG)

	self.__n_remove_triggered = false

end

function RewardBagSugar:on_destroy()


end

function RewardBagSugar:on_reset(c_msg)

	self.__n_remove_triggered = false

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")

end

function RewardBagSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function RewardBagSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function RewardBagSugar:on_start_remove(c_msg)

	if not c_msg.n_handled then

		self.__c_state_machine:do_event("remove")

	end


end

function RewardBagSugar:remove_in_range(remove_list)

	local result = false

	if not self:get_category():has("fenced") then

		local selector = self:get_component("selector")
		local msg = MessagePool:get_free_ex("select", {ban_message = true})
		selector:send_message(msg)

		local select_result = msg.x_result

		for i_s = 1, #select_result do

			for i_r = 1, #remove_list do

				result = (select_result[i_s] == remove_list[i_r])

				if result then

					return result

				end

			end

		end

	end

	return result

end

function RewardBagSugar:on_link_remove_start(c_msg)
-- print("start_ice_check")
	self.__n_remove_triggered = self:remove_in_range(c_msg.x_param)

end

function RewardBagSugar:on_trigger_indicate(c_msg)

	if self:remove_in_range(c_msg.x_param) then

		self:send_message(MessagePool:get_free_ex("highlight_on", "", Message.N_TUNNEL))

	else

		self:send_message(MessagePool:get_free_ex("highlight_off", "", Message.N_TUNNEL))


	end

end

function RewardBagSugar:on_link_remove_finish(c_msg)

	if self.__n_remove_triggered then

		self:send_message(MessagePool:get_free_ex("start_remove"))

	end

	self.__n_remove_triggered = false

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("self_remove_finish"))

end

function RewardBagSugar:on_remove_started()

	-- self:send_message(MessagePool:get_free_ex("finish_remove", "", Message.N_TUNNEL))

end

function RewardBagSugar:on_finish_remove()

	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))

	GameGlobals.Scene:send_message(MessagePool:get_free_ex("blast_finale", self, Message.N_TUNNEL))
	
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

end

function RewardBagSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function RewardBagSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function RewardBagSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},

	})
	self.__c_state_machine:add_transition({

		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end

function RewardBagSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end


RewardBagSugar.LISTENER_CONFIG = {
	
	{"select", {before = RewardBagSugar.on_select}},
	{"unselect", {before = RewardBagSugar.on_unselect}},
	{"start_remove", {after = RewardBagSugar.on_start_remove}},
	{"finish_remove", {before = RewardBagSugar.on_finish_remove}},
	{"remove_started", {before = RewardBagSugar.on_remove_started}},
	{"link_remove_start", {before = RewardBagSugar.on_link_remove_start}},
	{"link_remove_finish", {before = RewardBagSugar.on_link_remove_finish}},
	{"lightning_remove_begin", {before = RewardBagSugar.on_link_remove_start}},
	{"lightning_remove_finish", {before = RewardBagSugar.on_link_remove_finish}},
	{"trigger_indicate", {before = RewardBagSugar.on_trigger_indicate}},
	{"reset", {before = RewardBagSugar.on_reset}},

}

RewardBagSugar.PROPERTY_CONFIG = {
	
	{"material", cc.PhysicsMaterial(0.3, 0.0, 0.0)},
	{"radius", 60},
	{"state", "normal"},
	{"dynamic", true},
	{"id", "rewardbag"},

}


return RewardBagSugar