-- ## Imports ## --

local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local CircleCollider = require "GameEx.Components.Colliders.CircleCollider"
local TimingBombSugarVisual = require "GameEx.Components.Visuals.TimingBombSugarVisual"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"
local GameGlobals = require "GameEx.GameGlobals"
local ObjectFactory = require "GameEx.Foundations.ObjectFactory"
local ProbGetter = require "GameEx.Foundations.Utils.ProbGetter"
local RandomGenerator = require "GameEx.Foundations.Utils.RandomGenerator"

-- ## Imports ## --




local TimingBombSugar = class("TimingBombSugar", function() return GameObject:create() end)



function TimingBombSugar:ctor()

	self.__c_prob_getter = ProbGetter:create(GameGlobals.RandomGenerator)

	self:init_state_machine()

	local visual = ObjectFactory:get_instance():get("TimingBombSugarVisual")
	self:add_component(visual, "visual")

	local collider = ObjectFactory:get_instance():get("CircleCollider")
	self:add_component(collider, "collider")

	local category = self:get_category()
	category:add("timing_bomb")

	self.__n_random_flag = false

	self:register_listeners_with_config(TimingBombSugar.LISTENER_CONFIG)
	self:register_properties_with_config(TimingBombSugar.PROPERTY_CONFIG)


end

function TimingBombSugar:on_enter()

	if self.__n_random_flag then

		self:set("color", self.__c_prob_getter:get())

	end

	self.__n_random_flag = false

end

function TimingBombSugar:on_reset(c_msg)

	local default_config = c_msg.x_param

	for key, value in pairs(default_config) do

		self:set(key, value)

	end

	self:set("state", "normal")

end

function TimingBombSugar:on_lock()
	
	self.__c_state_machine:do_event("lock")

end

function TimingBombSugar:on_select()
	
	self.__c_state_machine:do_event("select")

end

function TimingBombSugar:on_unselect()

	self.__c_state_machine:do_event("unselect")

end

function TimingBombSugar:on_start_remove(c_msg)

	if not c_msg.n_handled then

		self.__c_state_machine:do_event("remove")

	end

end

function TimingBombSugar:on_remove_started(c_msg)

	self:send_message(MessagePool:get_free_ex("finish_remove", "", Message.N_TUNNEL))

end

function TimingBombSugar:on_finish_remove()
-- 
	self:send_message(MessagePool:get_free_ex("remove_finished", "", Message.N_TUNNEL))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("remove_sugar", self, Message.N_TUNNEL))

end

function TimingBombSugar:on_player_round_finish(c_msg)

	if c_msg.x_param == "player" then

		self.__t_current_count_down = self.__t_current_count_down - 1

		self:send_message(MessagePool:get_free_ex("reduce_count", "", Message.N_TUNNEL))

		if self.__t_current_count_down == 0 then

			GameGlobals.Scene:send_message(MessagePool:get_free_ex("timing_bomb_explode"))

		end

	end

end

function TimingBombSugar:after_set_count_down(c_msg)

	self.__t_current_count_down = c_msg.x_param.x_new_value

end

function TimingBombSugar:after_get_count_down(c_msg)

	c_msg.x_param.x_return_value = self.__t_current_count_down

end

function TimingBombSugar:after_set_state(c_msg)

	self.__c_state_machine:do_transition(c_msg.x_param.x_new_value)

end

function TimingBombSugar:after_set_id(c_msg)

	self:setName(c_msg.x_param.x_new_value)

end

function TimingBombSugar:after_get_state(c_msg)

	c_msg.x_param.x_return_value = self.__c_state_machine:get_current_state()

end

function TimingBombSugar:before_set_color(c_msg)

	if c_msg.x_param.x_new_value == "random" then

		self.__n_random_flag = true

	else

		self.__n_random_flag = false

	end

end

function TimingBombSugar:after_set_prob_red(c_msg)

	self:change_prob("red", c_msg.x_param.x_new_value)

end

function TimingBombSugar:after_set_prob_green(c_msg)

	self:change_prob("green", c_msg.x_param.x_new_value)

end

function TimingBombSugar:after_set_prob_yellow(c_msg)

	self:change_prob("yellow", c_msg.x_param.x_new_value)

end

function TimingBombSugar:after_set_prob_pink(c_msg)

	self:change_prob("pink", c_msg.x_param.x_new_value)

end

function TimingBombSugar:after_set_prob_blue(c_msg)

	self:change_prob("blue", c_msg.x_param.x_new_value)

end

function TimingBombSugar:after_set_prob_white(c_msg)

	self:change_prob("white", c_msg.x_param.x_new_value)

end

function TimingBombSugar:change_prob(s_key, n_prob)

	if self.__c_prob_getter:has(s_key) then

		self.__c_prob_getter:modify(s_key, n_prob)

	else

		self.__c_prob_getter:add(s_key, s_key, n_prob)

	end	

end

function TimingBombSugar:init_state_machine()

	function on_enter_remove()

		self:send_message(MessagePool:get_free_ex("remove_started", "", Message.N_TUNNEL))

	end

	function on_enter_select()

		self:send_message(MessagePool:get_free_ex("selected", "", Message.N_TUNNEL))

	end

	function on_leave_select()
		
		self:send_message(MessagePool:get_free_ex("unselected", "", Message.N_TUNNEL))

	end


	self.__c_state_machine = StateMachine:create()
	self.__c_state_machine:add_state({

		{name = "normal", enter = on_enter_normal},
		{name = "select", enter = on_enter_select, leave = on_leave_select},
		{name = "remove", enter = on_enter_remove},
		{name = "lock"},

	})
	self.__c_state_machine:add_transition({

		{event = "lock", from = "normal", to = "lock"},
		{event = "remove", from = "normal", to = "remove"},
		{event = "remove", from = "select", to = "remove"},
		{event = "remove", from = "lock", to = "remove"},
		{event = "select", from = "normal", to = "select"},
		{event = "select", from = "lock", to = "select"},
		{event = "unselect", from = "select", to = "normal"},

	})
	self.__c_state_machine:do_set("normal")

end

function TimingBombSugar:after_set_tagged(c_msg)

	if c_msg.x_param.x_new_value == true then

		self:get_category():add("tagged")

	else

		self:get_category():remove("tagged")

	end

end


TimingBombSugar.LISTENER_CONFIG = {

	{"lock", {before = TimingBombSugar.on_lock}},
	{"select", {before = TimingBombSugar.on_select}},
	{"unselect", {before = TimingBombSugar.on_unselect}},
	{"start_remove", {after = TimingBombSugar.on_start_remove}},
	{"finish_remove", {before = TimingBombSugar.on_finish_remove}},
	{"remove_started", {before = TimingBombSugar.on_remove_started}},
	{"round_finish", {before = TimingBombSugar.on_player_round_finish}},
	{"reset", {before = TimingBombSugar.on_reset}},
	{"enter", {before = TimingBombSugar.on_enter}},

}

TimingBombSugar.PROPERTY_CONFIG = {
	
	{"dynamic", true},
	{"radius", 62},
	{"state", "normal"},
	{"count_down", 11},
	{"prob_red", 0},
	{"prob_green", 0},
	{"prob_blue", 0},
	{"prob_yellow", 0},
	{"prob_pink", 10},
	{"prob_white", 0},
	{"color", "red"},
	{"id", "timingbomb"},
	{"tagged", "false"},

}


return TimingBombSugar