-- ## Imports ## --

local GameGlobals = require "GameEx.GameGlobals"
local GameObject = require "GameEx.Foundations.GameObject"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local ResourceManager = require "GameEx.Foundations.ResourceManager"
local ResourceManager_Core = require "core.ResourceManager"
-- ## Imports ## --

local FanSugarVisual = class("FanSugarVisual", function() return GameObject:create() end)

function FanSugarVisual:ctor()

	local spine = ResourceManager:get_instance():get_resource_by_key("fan_sugar_spine", true)
	spine:clearTracks()
	spine:setOpacity(255)
	spine:setRotation(0)
	spine:setScale(1.0)
	spine:setLocalZOrder(2)
	spine:setAnimation(0, "center", false)
	self:addChild(spine, 9,9)

	self.__t_leaves = {}
	self.__t_current_leaf = 0

	self:register_listeners_with_config(FanSugarVisual.LISTENER_CONFIG)
	self:register_properties_with_config(FanSugarVisual.PROPERTY_CONFIG)

end

function FanSugarVisual:on_destroy()

	self:removeChildByTag(9)

	for i = 1, #self.__t_leaves do

		self.__t_leaves[i]:removeFromParent()

	end

end

function FanSugarVisual:on_fill_color(c_msg)

	self.__t_current_leaf = self.__t_current_leaf + 1

	if self.__t_current_leaf <= #self.__t_leaves then

		local skin_name = string.format("%s_%d", c_msg.x_param, self:get("volume"))

		self.__t_leaves[self.__t_current_leaf]:setSkin(skin_name)
		self.__t_leaves[self.__t_current_leaf].skin_name = skin_name

	end

	local volume = self:get("volume")
	local delta_angle = 360 / volume
	local angle_offset = 0

	if volume == 4 then

		angle_offset = 45

	end

	local spine = ResourceManager:get_instance():get_resource_by_key("fan_sugar_spine")
	spine:clearTracks()
	spine:setOpacity(255)
	spine:setRotation(0)
	spine:setScale(self:getScale())
	spine:setLocalZOrder(1)
	spine:setSkin(self.__t_leaves[self.__t_current_leaf].skin_name)
	spine:setRotation(angle_offset + (self.__t_current_leaf) * delta_angle)
	spine:setPosition(cc.p(self.__c_parent:getPosition()))
	spine:setAnimation(0, "roll", true)
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", spine))

	local fade = cc.FadeTo:create(0.4, 0)
	local scale = cc.ScaleTo:create(0.4, 1.4)
	local sequence = cc.Sequence:create(cc.DelayTime:create(0.5), cc.RemoveSelf:create())
	local spawn = cc.Spawn:create(fade, scale, sequence)
	spine:runAction(spawn)

end

function FanSugarVisual:on_assemble()

	self.__c_parent:addChild(self, 1)

end

function FanSugarVisual:on_disassemble()

	self.__c_parent:removeChild(self)

end

function FanSugarVisual:after_set_volume(c_msg)

	for i = 1, #self.__t_leaves do

		self.__t_leaves[i]:removeFromParent()

	end

	local volume = c_msg.x_param.x_new_value
	local delta_angle = 360 / volume
	local angle_offset = 0

	if volume == 4 then

		angle_offset = 45

	end

	for i = 1, volume do

		local skin_name = string.format("none_%d", volume)

		local spine = ResourceManager:get_instance():get_resource_by_key("fan_sugar_spine")
		spine:clearTracks()
		spine:setOpacity(255)
		spine:setScale(1.0)
		spine:setLocalZOrder(1)
		spine:setSkin(skin_name)
		spine:setRotation(angle_offset + (i) * delta_angle)
		spine:setAnimation(0, "roll", true)
		spine:setPosition(cc.p(0, 0))
		self:addChild(spine)

		self.__t_leaves[i] = spine

	end

end

function FanSugarVisual:on_remove_started()
	MusicManager:getInstance().play_effect(ResourceManager_Core:getInstance():getAudioPathByKey("Fan_Audio"), false)

	local particle = ResourceManager:get_instance():get_resource_by_key("fan_remove_particle")
	particle:setAutoRemoveOnFinish(true)
	particle:setLocalZOrder(9)
	particle:setScale(1.3 * self:getScale())
	particle:setPosition(cc.p(self.__c_parent:getPosition()))
	GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", particle))

	local volume = self:get("volume")
	local delta_angle = 360 / volume
	local angle_offset = 0

	if volume == 4 then

		angle_offset = 45

	end

	for i = 1, #self.__t_leaves do

		local spine = ResourceManager:get_instance():get_resource_by_key("fan_sugar_spine")
		spine:clearTracks()
		spine:setOpacity(255)
		spine:setScale(self:getScale())
		spine:setLocalZOrder(1)
		spine:setSkin(self.__t_leaves[i].skin_name)
		spine:setRotation(angle_offset + (i) * delta_angle)
		spine:setPosition(cc.p(self.__c_parent:getPosition()))
		spine:setAnimation(0, "roll", true)
		GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", spine))


		local rotation = cc.RotateBy:create(1.5, 1080)
		local ease = cc.EaseSineIn:create(rotation)
		local delay = cc.DelayTime:create(0.5 + 0.099 * i)
		local call_func = cc.CallFunc:create(function()

			spine:setAnimation(0, "go", false)
			spine:registerSpineEventHandler(function (event)

				spine:registerSpineEventHandler(function(event) end, sp.EventType.ANIMATION_END)
				spine:removeFromParent()

			end, sp.EventType.ANIMATION_END)			

		end)
		local sequence = cc.Sequence:create(delay, call_func)
		local spawn = cc.Spawn:create(ease, sequence)
		spine:runAction(spawn)

	end

end

function FanSugarVisual:after_set_radius(c_msg)

	self:setScale(c_msg.x_param.x_new_value / 80)

end

function FanSugarVisual:on_reset(c_msg)

	self.__t_current_leaf = 0

	for i = 1, #self.__t_leaves do

		self.__t_leaves[i]:removeFromParent()

	end	

end

FanSugarVisual.LISTENER_CONFIG = {

	{"start_remove", {before = FanSugarVisual.on_start_remove}},
	{"append_to", {before = FanSugarVisual.on_append_to}},
	{"assemble", {before = FanSugarVisual.on_assemble}},
	{"disassemble", {before = FanSugarVisual.on_disassemble}},
	{"fill_color", {before = FanSugarVisual.on_fill_color}},
	{"remove_started", {before = FanSugarVisual.on_remove_started}},
	{"destroy", {before = FanSugarVisual.on_destroy}},
	{"reset", {before = FanSugarVisual.on_reset}},

}

FanSugarVisual.PROPERTY_CONFIG = {
	
	{"volume", 0},
	{"radius", 60}

}

return FanSugarVisual