-- ## Imports ## --

local GameGlobals = require "GameEx.GameGlobals"
local GameObject = require "GameEx.Foundations.GameObject"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local UtilFunction = require "GameEx.Foundations.Utils.Function"
local ResourceManager = require "GameEx.Foundations.ResourceManager"
local ResourceManager_Core = require "core.ResourceManager"

-- ## Imports ## --

local TransformSugarVisual = class("TransformSugarVisual", function() return GameObject:create() end)

function TransformSugarVisual:ctor()

	self.__c_spine = ResourceManager:get_instance():get_resource_by_key("transform_sugar_spine", true)
	self.__c_spine:setAnimation(0, "ding", false)
	self:addChild(self.__c_spine)

	self:register_listeners_with_config(TransformSugarVisual.LISTENER_CONFIG)
	self:register_properties_with_config(TransformSugarVisual.PROPERTY_CONFIG)

end

function TransformSugarVisual:on_destroy()

	self.__c_spine:removeFromParent()

end

function TransformSugarVisual:on_reset()

	self.__c_spine:setAnimation(0, "untouch", false)
	self.__c_spine:addAnimation(0, "ding", false)

end

function TransformSugarVisual:on_assemble()

	self.__c_parent:addChild(self)

end

function TransformSugarVisual:on_disassemble()

	self.__c_parent:removeChild(self)

end

function TransformSugarVisual:on_select()

	-- self:set_skin(self:get("color"), "_face")

	self.__c_spine:setAnimation(0, "barrier_114_touch", false)
	self.__c_spine:addAnimation(0, "barrier_114_hold", true)

	self.__n_original_z = self.__c_parent:getLocalZOrder()
	self.__c_parent:setLocalZOrder(999)

	self:setScale(1.0)

end

function TransformSugarVisual:on_unselect()

	self:set_skin(self:get("color"), "")

	self.__c_spine:setAnimation(0, "untouch", false)

	self.__c_parent:setLocalZOrder(self.__n_original_z)

	self:setScale(0.95)

end

function TransformSugarVisual:on_start_remove()
	MusicManager:getInstance().play_effect(ResourceManager_Core:getInstance():getAudioPathByKey("Ball_Audio"), false)
	
	self:set_skin(self:get("color"), "")

	local node_position = cc.p(self:getPosition())
	local world_position = self:getParent():convertToWorldSpace(node_position)

   	local particle_1 = ResourceManager:get_instance():get_resource_by_key("normal_remove_particle_1")
   	particle_1:setLocalZOrder(9)
   	particle_1:setPosition(world_position)
   	particle_1:setAutoRemoveOnFinish(true)
   	GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", particle_1))

   	local particle_2 = ResourceManager:get_instance():get_resource_by_key("normal_remove_particle_2")
   	particle_2:setLocalZOrder(9)
   	particle_2:setPosition(world_position)
   	particle_2:setAutoRemoveOnFinish(true)
   	GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", particle_2))

   	local particle_3 = ResourceManager:get_instance():get_resource_by_key("normal_remove_particle_3")
   	particle_3:setLocalZOrder(9)
   	particle_3:setPosition(world_position)
   	particle_3:setAutoRemoveOnFinish(true)
   	GameGlobals.Scene:send_message(MessagePool:get_free_ex("normal_add_child", particle_3))

end


function TransformSugarVisual:on_finish_remove()



end

function TransformSugarVisual:on_highlight_on()

	if self.on then return end
	self.on = true
	self.__c_spine:setAnimation(2, "barrier_114_tishi", true)

end

function TransformSugarVisual:on_highlight_off()

	self.on = false
	self.__c_spine:setAnimation(2, "barrier_114_tishi_end", false)

end

function TransformSugarVisual:after_set_radius(c_msg)

	self:setScale(c_msg.x_param.x_new_value / 64)

end

function TransformSugarVisual:after_set_color(c_msg)

	self:set_skin(c_msg.x_param.x_new_value, "")

end

function TransformSugarVisual:set_skin(color, suffix)

	local str = string.format("%s%s", color, suffix)

	self.__c_spine:setSkin(str)

end

function TransformSugarVisual:on_start_transform()

	self.__c_spine:setAnimation(0, "barrier_114", false)

	local delay = cc.DelayTime:create(0.2)
	local call = cc.CallFunc:create(function()

		self.__c_parent:send_message(MessagePool:get_free_ex("finish_transform"))

	end)
	local sequence = cc.Sequence:create(delay, call)

	self:runAction(sequence)

end

TransformSugarVisual.LISTENER_CONFIG = {

	{"assemble", {before = TransformSugarVisual.on_assemble}},
	{"disassemble", {before = TransformSugarVisual.on_disassemble}},
	{"selected", {before = TransformSugarVisual.on_select}},
	{"unselected", {before = TransformSugarVisual.on_unselect}},
	{"remove_started", {before = TransformSugarVisual.on_start_remove}},
	{"remove_finished", {before = TransformSugarVisual.on_finish_remove}},
	{"highlight_on", {before = TransformSugarVisual.on_highlight_on}},
	{"highlight_off", {before = TransformSugarVisual.on_highlight_off}},
	{"reset", {before = TransformSugarVisual.on_reset}},
	{"start_transform", {before = TransformSugarVisual.on_start_transform}},
	{"destroy", {before = TransformSugarVisual.on_destroy}},

}

TransformSugarVisual.PROPERTY_CONFIG = {
	
	{"radius", 1},
	{"color", "red"}

}

return TransformSugarVisual