-- ## Imports ## --

local UtilFunction = require "GameEx.Foundations.Utils.Function"
local Message = require "GameEx.Foundations.Message"
local MessagePool = require "GameEx.Foundations.MessagePool"
local Category = require "GameEx.Foundations.Category"
local CategoryFilter = require "GameEx.Components.Filters.CategoryFilter"
local PropertyFilter = require "GameEx.Components.Filters.PropertyFilter"

-- ## Imports ## --

local GameObject = class("GameObject", function() return cc.Node:create() end)

function GameObject:ctor()

	self.__t_named_components = {}
	self.__t_named_properties = {}
	self.__t_event_listeners = {}
	self.__c_category = Category:create()
	self.__c_category_filter = CategoryFilter:create(self.__c_category)
	self.__c_property_filter = PropertyFilter:create(self)
	self.__c_parent = nil
	self.__c_root = nil
	self.__n_max_prior = 0

	self:set_event_listener("get", {before = self.before_get, after = self.after_get})
	self:set_event_listener("set", {before = self.before_set, after = self.after_set})

	self:add_property("category", "none")

end 

function GameObject:destroy()

	self:remove_all_components()

	local msg = MessagePool:get_free()
	msg.c_sender = self
	msg.s_name = "destroy"

	self:send_message(msg)

end

function GameObject:get_category()

	return self.__c_category

end

function GameObject:add_component(c_component, s_name, priority)

	c_component:retain()
	c_component.__c_parent = self

	local msg = MessagePool:get_free()
	msg.c_sender = self
	msg.s_name = "assemble"
	msg.e_transmit = Message.N_DIRECT

	c_component:send_message(msg)

	local _prior = priority or 0

	if _prior > self.__n_max_prior then self.__n_max_prior = _prior end

	c_component._prior = _prior

	self.__t_named_components[_prior] = self.__t_named_components[_prior] or {}

	self.__t_named_components[_prior][s_name] = c_component
	self.__t_named_components[s_name] = c_component

end

function GameObject:remove_component(s_name)

	local component = self.__t_named_components[s_name]

	if component then

		local msg = MessagePool:get_free()
		msg.c_sender = self
		msg.s_name = "disassemble"
		msg.e_transmit = Message.N_DIRECT

		component:send_message(msg)

		self.__t_named_components[s_name] = nil
		self.__t_named_components[component._prior][s_name] = nil

		component.__c_parent = nil
		component:release()

	end

end

function GameObject:get_component(s_name)

	return self.__t_named_components[s_name]

end

function GameObject:add_property(s_name, x_default_value)

	local property_holder = {}
	property_holder.s_before_getter_name = string.format("before_get_%s", s_name)
	property_holder.s_before_setter_name = string.format("before_set_%s", s_name)
	property_holder.s_after_getter_name = string.format("after_get_%s", s_name)
	property_holder.s_after_setter_name = string.format("after_set_%s", s_name)

	self.__t_named_properties[s_name] = property_holder

	self:set(s_name, x_default_value)

end

function GameObject:remove_property(s_name)

	self.__t_named_properties[s_name] = nil

end

function GameObject:get_property(s_name)

	return self.__t_named_properties[s_name]

end

function GameObject:get(s_name)

	local msg = MessagePool:get_free()
	msg.s_name = "get"
	msg.x_param = {s_property_name = s_name}
	msg.c_sender = self
	msg.e_transmit = Message.N_TUNNEL

	self:send_message(msg)

	return msg.x_param.x_return_value

end

function GameObject:set(s_name, x_value)

	local msg = MessagePool:get_free()
	msg.s_name = "set"
	msg.x_param = {s_property_name = s_name, x_new_value = x_value}
	msg.c_sender = self
	msg.e_transmit = Message.N_TUNNEL

	self:send_message(msg)

end

function GameObject:before_get(c_msg)

	local property = self:get_property(c_msg.x_param.s_property_name)

	if property then

		c_msg.x_param.x_return_value = property.x_value

		local before_getter = property.s_before_getter_name

		if self[before_getter] then

			self[before_getter](self, c_msg)

		end

	end

end

function GameObject:before_set(c_msg)

	local property = self:get_property(c_msg.x_param.s_property_name)

	if property then

		local before_setter = property.s_before_setter_name

		if self[before_setter] then

			self[before_setter](self, c_msg)

		end

	end

end

function GameObject:after_get(c_msg)

	local property = self:get_property(c_msg.x_param.s_property_name)

	if property then

		local after_getter = property.s_after_getter_name

		if self[after_getter] then

			self[after_getter](self, c_msg)

		end

	end

end

function GameObject:after_set(c_msg)

	local property = self:get_property(c_msg.x_param.s_property_name)

	if property then

		local after_setter = property.s_after_setter_name

		if self[after_setter] then

			self[after_setter](self, c_msg)

		end

		property.x_value = c_msg.x_param.x_new_value

	end

end

function GameObject:send_message(c_message)

	self:retain()

	local listeners =  self.__t_event_listeners[c_message.s_name]

	if listeners and listeners.before then

		listeners.before(self, c_message)

	end

	if c_message.e_transmit == Message.N_TUNNEL then

		for i = 0, self.__n_max_prior do 

			local components = self.__t_named_components[i]

			if components then

				for name, component in pairs(components) do

					component:send_message(c_message)

				end				

			end

		end

	end

	if listeners and listeners.after then

		listeners.after(self, c_message)

	end

	self:release()

end

function GameObject:set_event_listener(s_event_name, t_listener)

	self.__t_event_listeners[s_event_name] = t_listener

end

function GameObject:get_event_listener(s_event_name)

	return self.__t_event_listeners[s_event_name]

end

function GameObject:remove_event_listener(s_event_name)

	self.__t_event_listeners[s_event_name] = nil

end

function GameObject:remove_all_components()

	for name, component in pairs(self.__t_named_components) do

		if type(name) == "string" then

			self:remove_component(name)

		end

	end

end

function GameObject:register_listeners_with_config(t_config)

	local config_size = #t_config

	for i = 1, config_size do

		local config_item = t_config[i]

		self:set_event_listener(config_item[1], config_item[2])

	end

end

function GameObject:register_properties_with_config(t_config)

	local config_size = #t_config

	for i = 1, config_size do

		local config_item = t_config[i]

		self:add_property(config_item[1], UtilFunction:copy(config_item[2]))

	end

end

function GameObject:property_filter(t_filter_config)

	return self.__c_property_filter:filter(t_filter_config)

end

function GameObject:category_filter(t_filter_config)

	return self.__c_category_filter:filter(t_filter_config)

end

function GameObject:after_set_category(c_msg)

	local category = self:get_category()
	category:add(c_msg.x_param.x_new_value)

end

function GameObject:getPositionWorld()

	local node_pos = cc.p(self:getPosition())
	local world_pos = self:getParent():convertToWorldSpace(node_pos)

	return world_pos.x, world_pos.y

end

GameObject.C_EMPTY_COMPONENT = GameObject:create()

return GameObject
