local UtilFunction = require "GameEx.Foundations.Utils.Function"
local ConfigManager = require "GameEx.Foundations.ConfigManager"

local TileMapParser = class("Tile", {})

function TileMapParser:ctor()

	self.__t_parse_result = {}
	self.__t_file_content = {}

	self.__t_position_offset = {}
	self.__t_position_offset.x = 40
	self.__t_position_offset.y = 220

end

function TileMapParser:parse(s_file_path, level_index)

	package.loaded[s_file_path] = nil

	self.__n_level_index = level_index
	self.__t_file_content = require(s_file_path)
	self.__t_parse_result = {}

	self:get_config()
	self:get_canvas()

	self:remove_inactived_config()

	return self.__t_parse_result

end

function TileMapParser:get_config()

	self.__t_parse_result.config = {}

	self:get_sugar_config()
	self:get_world_config()
	self:parse_ref()
	self:parse_boss_config()
	self:parse_span_config(self.__n_level_index)

end

function TileMapParser:get_world_config(t_file_content, t_config_set)

	local config = self:get_or_create_config("GameScene")
	config.actived = true
	config.properties = self.__t_file_content.properties
	config.properties["column"] = self.__t_file_content.width
	config.properties["row"] = self.__t_file_content.height
	config.properties["tile_width"] = self.__t_file_content.tilewidth
	config.properties["tile_height"] = self.__t_file_content.tileheight

end

function TileMapParser:parse_boss_config()

	local config = self:get_or_create_config("GameScene")

	if config.properties.has_boss then

		local boss_properties = require(string.format("GameEx.Resources.Levels.boss_%d", self.__n_level_index))
		local boss_config = self:get_or_create_config("BossSugar")
		boss_config.actived = true
		boss_config.properties["max_hp"] = boss_properties["max_hp"]
		boss_config.properties["texture"] = boss_properties["texture"]
		boss_config.properties["hp_stage_config"] = boss_properties["hp_stage_config"]

		for i = 1, #boss_properties["hp_stage_config"] do

			local attack_type = boss_properties["hp_stage_config"][i].config.attack_config.type
			self:recursive_check_tile(attack_type)

		end

	end

end

function TileMapParser:get_sugar_config()

	self:get_from_tileset()
	self:get_from_objectset()

end

function TileMapParser:get_or_create_config(key)

	self.__t_parse_result.config[key] = self.__t_parse_result.config[key] or {name = key, properties = {}, actived = false, tiles = {}}

	return self.__t_parse_result.config[key]

end

function TileMapParser:get_from_tileset()

	for _, tile_set in ipairs(self.__t_file_content.tilesets) do

		local config = self:get_or_create_config(tile_set.name)
		config.properties = tile_set.properties or {}
		config.actived = config.properties["actived"]
		config.tiles = tile_set.tiles

	end

	self:get_or_create_config("ColorfulSugar").actived = true
	self:get_or_create_config("CrossSugar").actived = true
	self:get_or_create_config("BombSugar").actived = true
	self:get_or_create_config("LightningSugar").actived = true
	self:get_or_create_config("LineSugar").actived = true

end

function TileMapParser:get_from_objectset()

	for _, layer in ipairs(self.__t_file_content.layers) do

		if layer.type == "objectgroup" then

			local config = self:get_or_create_config(layer.name)
			config.properties = layer.properties
			config.actived = true

		end

	end

end

function TileMapParser:get_canvas(t_file_content)

	self.__t_parse_result.canvas = {}

	for _, layer in ipairs(self.__t_file_content.layers) do

		if layer.type == "tilelayer" then

			self:get_from_tile_layer(layer)

		elseif layer.type == "objectgroup" then

			self:get_from_object_layer(layer)

		end

	end

	return canvas

end

function TileMapParser:get_from_tile_layer(t_layer_content)

	local column_number = self.__t_parse_result.config.GameScene.properties.column
	local row_number = self.__t_parse_result.config.GameScene.properties.row 
	local width = self.__t_parse_result.config.GameScene.properties.tile_width
	local height = self.__t_parse_result.config.GameScene.properties.tile_height
	local layer_offset_x = t_layer_content.offsetx
	local layer_offset_y = -t_layer_content.offsety

	for row = 1, row_number do

		for column = 1, column_number do

			local index = t_layer_content.data[ (row_number - row) * column_number + column ]

			for _, tile_set in ipairs(self.__t_file_content.tilesets) do

				if index >= tile_set.firstgid and index < tile_set.firstgid + tile_set.tilecount then

					local tile_index = index - tile_set.firstgid

					self:get_or_create_config(tile_set.name).actived = true

					local config = {}
					config.name = tile_set.name
					config.position = {}
					config.position.x = width / 2 + width * (column - 1) + self.__t_position_offset.x + layer_offset_x
					config.position.y = height / 2 + height * (row - 1) + self.__t_position_offset.y + layer_offset_y
					config.properties = {}
					self.__t_parse_result.canvas[#self.__t_parse_result.canvas + 1] = config

					for _, tile in ipairs(tile_set.tiles) do

						if tile.id == tile_index then

							config.properties = tile.properties

						end

					end

					self:recursive_check_tile(tile_set.name, tile_index)

					break

				end

			end

		end

	end

end

function TileMapParser:get_from_object_layer(t_layer_content)

	local column_number = self.__t_parse_result.config.GameScene.properties.column
	local row_number = self.__t_parse_result.config.GameScene.properties.row
	local width = self.__t_parse_result.config.GameScene.properties.tile_width
	local height = self.__t_parse_result.config.GameScene.properties.tile_height

	for _, object in ipairs(t_layer_content.objects) do 

		self:get_or_create_config(t_layer_content.name).actived = true

		if object.shape == "rectangle" then

			local config = {}
			config.name = t_layer_content.name
			config.position = {}
			config.position.x = object.x + (object.width * math.cos(math.rad(object.rotation)) - object.height * math.sin(math.rad(object.rotation))) / 2 + self.__t_position_offset.x
			config.position.y = height * row_number - (object.y + (object.width * math.sin(math.rad(object.rotation)) + object.height * math.cos(math.rad(object.rotation))) / 2) + self.__t_position_offset.y
			config.properties = object.properties
			config.properties["rotation"] = object.rotation
			config.properties["width"] = object.width
			config.properties["height"] = object.height

			self.__t_parse_result.canvas[#self.__t_parse_result.canvas + 1] = config

		elseif object.shape == "ellipse" then

			local config = {}
			config.name = t_layer_content.name
			config.position = {}
			config.position.x = object.x + object.width / 2 + self.__t_position_offset.x
			config.position.y = height * row_number - (object.y + object.width / 2) + self.__t_position_offset.y
			config.properties = object.properties
			config.properties["radius"] = object.width

			self.__t_parse_result.canvas[#self.__t_parse_result.canvas + 1] = config

		elseif object.shape == "polyline" then

			local config = {}
			config.name = t_layer_content.name
			config.position = {}
			config.position.x = object.x + self.__t_position_offset.x
			config.position.y = height * row_number - object.y + self.__t_position_offset.y
			config.properties = object.properties
			config.properties["rotation"] = object.rotation
			config.properties["edges"] = self:reverse_polyline_coordinate(object.polyline)

			self.__t_parse_result.canvas[#self.__t_parse_result.canvas + 1] = config

		end

	end

end

function TileMapParser:reverse_polyline_coordinate(t_polyline)

	local result = UtilFunction:copy(t_polyline)

	local max_y = -math.huge

	for i = 1, #result do

		if result[i].y > max_y then max_y = result[i].y end

	end

	for i = 1, #result do

		result[i].y = -result[i].y

	end

	return result

end

function TileMapParser:remove_inactived_config()

	for key, value in pairs(self.__t_parse_result.config) do

		if not value.actived then
			
			self.__t_parse_result.config[key] = nil

		end

	end

end

function TileMapParser:parse_span_config(n_level_id)

	local span_config_raw = nil

	pcall(function() span_config_raw = require(string.format("GameEx.Resources.Levels.span_%d", n_level_id)) end)

	if span_config_raw and self:span_config_validation(span_config_raw) then

		for i = 1, #span_config_raw do

			local span = span_config_raw[i]

			span["duration"] = span["duration"] or "INFINITTY"
			span["normal"] = self:build_normal_span(span["normal"])
			span["special"] = self:build_special_span(span["special"])

		end

		local config = self:get_or_create_config("AdvancedEmitter")
		config.actived = true
		config.properties["span"] = span_config_raw

		self.__t_parse_result.has_span = true

	end

end


function TileMapParser:build_normal_span(raw_data)

	local normal_span = raw_data or {}

	normal_span["normal_prob"] = normal_span["normal_prob"] or 0
	normal_span["normal_config"] = normal_span["normal_config"] or {}

	return normal_span

end

function TileMapParser:build_special_span(raw_data)

	local special_span = raw_data or {}

	special_span["special_prob"] = special_span["special_prob"] or 0
	special_span["special_number_config"] = special_span["special_number_config"] or {}
	special_span["special_config"] = special_span["special_config"] or {}

	local special_configs = special_span["special_config"]

	for i = 1, #special_configs do

		local config = special_configs[i]
		config["prob"] = config["prob"] or 0
		config["config_index"] = config["config_index"] or 0

		self:recursive_check_tile(config["type"], config["config_index"])

	end

	return special_span

end

function TileMapParser:span_config_validation(raw_config)

	for index, span in ipairs(raw_config) do

		ERROR_ASSERT( 
			span["special"] 
			and span["special"]["special_config"]
			and #(span["special"]["special_config"]) > 0 , "ERROR! Special config must have value !") 


		ERROR_ASSERT( 
			span["special"] 
			and span["special"]["special_config"]
			and (function() 

				local result = true

				for key, config in ipairs(span["special"]["special_config"]) do

					result = result and config["type"] and ConfigManager:get_instance():get_config(config["type"])

				end

				return result

			end)() , "ERROR! Special config must have available type !") 

	end

	return true

end

function TileMapParser:recursive_check_tile(class, tile_index)

	if not class then return end

	local config = self:get_or_create_config(class)
	config.actived = true

	for _, tile in ipairs(config.tiles) do

		if tile.id == tile_index and tile.properties["ref"] then

			for _, ref in ipairs(tile.properties["ref"]) do

				self:recursive_check_tile(ref.type, ref.tile_index)

			end

		end

	end

end

function TileMapParser:parse_ref()

	for _, config in pairs(self.__t_parse_result.config) do

		for _, tile in ipairs(config.tiles) do

			local config_tile = tile["properties"]

			config_tile["ref"] = {}

			local ref_number = config_tile["ref_number"] or 0

			for i = 1, ref_number do

				local class_type = string.format("ref_type_%d", i)
				local tile_index = string.format("ref_tile_index_%d", i)
				local prob = string.format("ref_prob_%d", i)

				config_tile["ref"][i] = { ["type"] = config_tile[class_type], ["tile_index"] = config_tile[tile_index], ["prob"] = config_tile[prob] }
				config_tile[class_type] = nil
				config_tile[tile_index] = nil
				config_tile[prob] = nil

			end

			config_tile["ref_number"] = nil			

		end

	end
	
end

return TileMapParser