local Downloader = require "core.Downloader"
local Decompressor = require "core.Decompressor"
local FileManipulator = require "GameEx.Foundations.Utils.FileManipulator"
local StateMachine = require "GameEx.Foundations.Utils.StateMachineRevised"

local UpdateManager = class("UpdateManager")

UpdateManager.TASK_STATE = {
	
	["RUNNING"] = 0,
	["COMPLETED"] = 1,

}

UpdateManager.RESULT_CODE = {
	
	["NO_RESULT"] = 0,

	["CURRENT_VERSION_INVALID"] = 1,
	["CURRENT_VERSION_UP_TO_DATE"] = 2,
	["CURRENT_VERSION_EXPIRED"] = 3,

	["RETRIEVE_VERSION_SUCCEEDED"] = 6,
	["RETRIEVE_VERSION_MANIFEST_INVALID"] = 4,
	["RETRIEVE_VERSION_FAILED"] = 5,

	["RETRIEVE_UPDATE_SUCCEEDED"] = 6,
	["RETRIEVE_UPDATE_NO_VALID_VERSION"] = 7,
	["RETRIEVE_UPDATE_FILE_INVALID"] = 8,
	["RETRIEVE_UPDATE_DECOMPRESS_FAILED"] = 9,
	["RETRIEVE_UPDATE_DOWNLOAD_SUCCEEDED"] = 10,
	["RETRIEVE_UPDATE_DOWNLOAD_FAILED"] = 11,

	["UPDATE_SUCCEEDED"] = 12,

}

UpdateManager.VERSION_MANIFEST_URL = ""
UpdateManager.UPDATE_FILE_URL = ""
UpdateManager.STORAGE_PATH = ""

function UpdateManager:ctor()

	-- Downloader下载器的实现在core/Downloader.lua文件中

	self.downloader = Downloader:create()

end

function UpdateManager:get_instance()

	if not UpdateManager.instance then

		UpdateManager.instance = UpdateManager:create()

	end

	return UpdateManager.instance

end

-- 初始化UpdateManager，必须在启动更新前进行调用，否则无法正常更新
--
--     manifest_url: 获取VERSION_MANIFEST.lua文件的URL (URL末尾需要包含‘/’)
--     file_url: 获取包含更新文件的zip包的URL
--     storage_path: 下载及更新文件的最终存放路径
--

function UpdateManager:init(manifest_url, file_url, storage_path)

	UpdateManager.VERSION_MANIFEST_URL = manifest_url
	UpdateManager.UPDATE_FILE_URL = file_url
	UpdateManager.STORAGE_PATH = storage_path

end

function UpdateManager:get_result_str(result_code)

	for code, str in ipairs(UpdateManager.RESULT_CODE) do

		if code == result_code then

			return str

		end

	end

end

-- 采用pcall的方式引入脚本，如果脚本不存在或无法解析则返回nil

function UpdateManager:safe_require(file_name)

	if pcall(function() require(file_name) end) then

		return require(file_name)

	else

		return nil

	end

end

-- 获取当前版本号
--
-- 游戏包的源码路径下需要包含一份current_version.lua的脚本文件来标示游戏包
-- 发布时所在的版本号，UpdateManager完成更新后将在存储路径下生成一份新的
-- current_version覆盖源码路径下的current_version来达到更新版本号的目的。

function UpdateManager:get_current_version()

	package.loaded["current_version"] = nil

	return self:safe_require("current_version")

end

-- 获取临时版本号
--
-- 临时版本号文件temp_version.lua包含在更新文件的zip包当中，表明当前zip包所
-- 在的版本号，UpdateManager更新完毕后会根据temp_version中的版本号在存储路
-- 径下生成一份新的current_version.lua文件，从而达到更新本地版本号的目的。

function UpdateManager:get_temp_version()

	package.loaded["temp_version"] = nil

	return self:safe_require("temp_version")

end

-- 获取版本明细
--
-- UpdateManager在更新开始前会从指定URL获取一份version_manifest.lua脚本文件
-- version_manifest中写明了当前的合法版本号列表，UpdateManager会将current_version
-- 中的版本号同version_manifest合法列表当中的版本号进行对比，如果当前版本号不存在
-- 则停止更新并返回错误，若当前版本号低于最新版本则从当前版本的下一版本开始按顺序
-- 下载新版本的zip包直至到达最新版本。

function UpdateManager:get_version_manifest()

	package.loaded["version_manifest"] = nil

	return self:safe_require("version_manifest")

end

-- 根据temp_version中的版本号在存储路径下生成新的current_version.lua文件

function UpdateManager:change_current_version()

	FileManipulator:getInstance():writeStringToFile("current_version", self:get_temp_version())

end

-- 删除更新相关文件（测试用）

function UpdateManager:clean()

	-- cc.FileUtils:getInstance():removeFile(string.format("%s%s.lua", UpdateManager.STORAGE_PATH, "version_manifest"))
	-- cc.FileUtils:getInstance():removeFile(string.format("%s%s.lua", UpdateManager.STORAGE_PATH, "temp_version"))
	-- cc.FileUtils:getInstance():removeFile(string.format("%s%s.lua", UpdateManager.STORAGE_PATH, "current_version"))

end

-- 比较版本号

function UpdateManager:compare_version()

	local version_manifest = self:get_version_manifest()
	local current_version = self:get_current_version()
	print("current_version:"..current_version)
if not version_manifest then JavaUtils.log("com.dayu", "No version manifest") end
if not current_version then JavaUtils.log("com.dayu", "No current version") end
	if version_manifest and current_version then
JavaUtils.log("com.dayu", tostring(#current_version))
		local current_stage = 0

		-- 将current_version中的版本号依序同version_manifest中的合法版本号进行比较
		-- version_manifest中的版本号按照新旧程度在数组中从低到高进行排列

		for stage, valid_version in ipairs(version_manifest) do
JavaUtils.log("com.dayu", valid_version)
			if valid_version == current_version then
JavaUtils.log("com.dayu", "version match")
				current_stage = stage
				break

			end 

		end

		if current_stage == 0 then -- 当前版本无效
JavaUtils.log("com.dayu", "Version not match")
			return UpdateManager.RESULT_CODE.CURRENT_VERSION_INVALID

		elseif current_stage == #version_manifest then -- 当前版本已是最新版本

			return UpdateManager.RESULT_CODE.CURRENT_VERSION_UP_TO_DATE

		else -- 当前版本不是最新版本

			return UpdateManager.RESULT_CODE.CURRENT_VERSION_EXPIRED

		end

	else -- 获取version_manifest或current_version失败，返回错误
JavaUtils.log("com.dayu", "Cannot get version")
		return UpdateManager.RESULT_CODE.CURRENT_VERSION_INVALID  

	end

end

-- 从version_manifest中获取相对当前版本的下一版本的版本号

function UpdateManager:get_next_version()

	local version_manifest = self:get_version_manifest()
	local current_version = self:get_current_version()

	if version_manifest and current_version then

		local current_stage = 0

		for stage, valid_version in ipairs(version_manifest) do

			if valid_version == current_version then

				current_stage = stage
				break

			end 

		end

		return version_manifest[current_stage + 1]

	else

		return nil

	end

end

-- 解压获取到的更新文件zip包
--
-- 注意：zip包中的文件的路径结构要与游戏资源的路径结构保持一致

function UpdateManager:decompress_update_file(task_info)

	task_info.task_type = "DECOMPRESS_UPDATE_FILE"

	local next_version = self:get_next_version()

	local file_name = string.format("%s%s.zip", UpdateManager.STORAGE_PATH, next_version)
	local decompress_succeeded = Decompressor.get_instance():decompress(file_name) -- Decompressor 是C提供的接口，实现部分在项目的source路径下（ZipUtilConnector）

	if decompress_succeeded then
JavaUtils.log("com.dayu", "Decompress Success")

		local temp_version = self:get_temp_version()
		local current_version = self:get_current_version()
		local next_version = self:get_next_version()

		if not temp_version then JavaUtils.log("com.dayu", "no temp version") end
		if not current_version then JavaUtils.log("com.dayu", "no current version") end

		if temp_version and current_version and (temp_version == next_version) then -- 比较temp_version与next_version，检查下载到的zip包是否正确
JavaUtils.log("com.dayu", "Update Success")
			self:change_current_version()

			task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
			task_info.task_progress = 100
			task_info.task_result_code = UpdateManager.RESULT_CODE.RETRIEVE_UPDATE_DOWNLOAD_SUCCEEDED

		else
JavaUtils.log("com.dayu", "version error")
			task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
			task_info.task_progress = 0
			task_info.task_result_code = UpdateManager.RESULT_CODE.RETRIEVE_UPDATE_FILE_INVALID

		end

	else
JavaUtils.log("com.dayu", "Decompress failed")
		task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
		task_info.task_progress = 0
		task_info.task_result_code = UpdateManager.RESULT_CODE.RETRIEVE_UPDATE_DECOMPRESS_FAILED

	end	

end

-- 获取version_manifest.lua文件
--
-- task_info为记录下载或更新任务状态的数据结构，UpdateManager会在执行更新的过程中
-- 主动更新task_info结构体中的数据，外部接口通过监测task_info结构体中的数据情况就可以
-- 获得当前更新任务的执行情况信息。

function UpdateManager:download_version_manifest(task_info)

	task_info.task_state = UpdateManager.TASK_STATE.RUNNING
	task_info.task_progress = 0
	task_info.task_result_code = UpdateManager.RESULT_CODE.NO_RESULT

	check_version_callbacks = {
		
		ON_SUCCESS = function(task_id) 

			if self:get_version_manifest() then

				JavaUtils.log("com.dayu", "Update Success:" .. "manifest")

				task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
				task_info.task_result_code = UpdateManager.RESULT_CODE.RETRIEVE_VERSION_SUCCEEDED

			else

				JavaUtils.log("com.dayu", "Update ERROR:" .. "manifest")

				task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
				task_info.task_result_code = UpdateManager.RESULT_CODE.VERSION_MANIFEST_INVALID

			end

		end,

		ON_FAILURE = function(err_code, err_str, task_id) 

			task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
			task_info.task_result_code = UpdateManager.RESULT_CODE.RETRIEVE_VERSION_MANIFEST_FAILED

			JavaUtils.log("com.dayu", "Update Error:" .. err_str)

		end,

		ON_PROGRESS = function(bytes_expected, bytes_recieved, task_id) 

			task_info.task_type = "RETRIEVE_VERSION_MANIFEST"
			task_info.task_progress = math.floor(bytes_recieved / bytes_expected * 100)

			JavaUtils.log("com.dayu", "Update Progress:" .. tostring(task_info.task_progress))

		end

	}

	task_param = {
		
		url = string.format("%s%s", UpdateManager.VERSION_MANIFEST_URL, "version_manifest.lua"),
		storage_path = string.format("%s%s.lua", UpdateManager.STORAGE_PATH, "version_manifest"),
		task_id = "check_version",
		on_success = check_version_callbacks.ON_SUCCESS,
		on_failure = check_version_callbacks.ON_FAILURE,
		on_progress = check_version_callbacks.ON_PROGRESS,

	}

	self.downloader:run_task(task_param)

end

-- 获取更新文件

function UpdateManager:download_update_file(task_info)

	local next_version = self:get_next_version()

	if next_version then

		task_info.task_state = UpdateManager.TASK_STATE.RUNNING
		task_info.task_progress = 0
		task_info.task_result_code = UpdateManager.RESULT_CODE.NO_RESULT

		get_file_callbacks = {
			
			ON_SUCCESS = function(task_id) 

				self:decompress_update_file(task_info)

			end,

			ON_FAILURE = function(err_code, err_str, task_id) 
JavaUtils.log("com.dayu", "Download File Failed")
				task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
				task_info.task_result_code = UpdateManager.RESULT_CODE.RETRIEVE_UPDATE_DOWNLOAD_FAILED

			end,

			ON_PROGRESS = function(bytes_expected, bytes_recieved, task_id) 

				task_info.task_type = "RETRIEVE_UPDATE_FILE"
				task_info.task_progress = math.floor(bytes_recieved / bytes_expected * 100)

			end,

		}

		task_param = {
			
			url = string.format("%s%s.zip", UpdateManager.UPDATE_FILE_URL, next_version),
			storage_path = string.format("%s%s.zip", UpdateManager.STORAGE_PATH, next_version),
			task_id = "get_update_file",
			on_success = get_file_callbacks.ON_SUCCESS,
			on_failure = get_file_callbacks.ON_FAILURE,
			on_progress = get_file_callbacks.ON_PROGRESS,

		}

		self.downloader:run_task(task_param)

	else

		task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
		task_info.task_progress = 0
		task_info.task_result_code = UpdateManager.RESULT_CODE.RETRIEVE_UPDATE_NO_VALID_VERSION

	end

end

-- 循环更新的实现，不要直接调用。

function UpdateManager:start_update(task_info_update)

	local compare_result = self:compare_version()

	if compare_result == UpdateManager.RESULT_CODE.CURRENT_VERSION_UP_TO_DATE then
JavaUtils.log("com.dayu", "Version Up To Date:")
		task_info_update.task_progress = 100
		task_info_update.task_state = UpdateManager.TASK_STATE.COMPLETED
		task_info_update.task_result_code = UpdateManager.RESULT_CODE.UPDATE_SUCCEEDED			

	elseif UpdateManager.RESULT_CODE.CURRENT_VERSION_EXPIRED then
JavaUtils.log("com.dayu", "Version Expired")
		task_info_update.task_progress = 0
		task_info_update.task_state = UpdateManager.TASK_STATE.RUNNING
		
		local task_info_getfile = {}

		self:download_update_file(task_info_getfile)

		local handler = 0

		handler = cc.Director:getInstance():getScheduler():scheduleScriptFunc(function()

			task_info_update.task_progress = task_info_getfile.task_progress

			if task_info_getfile.task_state == UpdateManager.TASK_STATE.COMPLETED then
	JavaUtils.log("com.dayu", "Task Finish:")
				cc.Director:getInstance():getScheduler():unscheduleScriptEntry(handler)
	JavaUtils.log("com.dayu", tostring(task_info_getfile.task_result_code))
				if task_info_getfile.task_result_code == UpdateManager.RESULT_CODE.RETRIEVE_UPDATE_DOWNLOAD_SUCCEEDED then
	JavaUtils.log("com.dayu", "Task Success")
					self:start_update(task_info_update) -- 递归调用start_update获取下一版本的更新文件
				else

					task_info_update.task_progress = 0
					task_info_update.task_state = UpdateManager.TASK_STATE.COMPLETED
					task_info_update.task_result_code = task_info_update.task_result_code

				end

			end

		end, 0.1, false)		

	end

end

-- 更新使用的功能函数。

function UpdateManager:schedule_subtask_observer(sub_task_info, task_info, success_condition)
JavaUtils.log("com.dayu", "Start Observer")
	
	local handler = 0

	handler = cc.Director:getInstance():getScheduler():scheduleScriptFunc(function()

		task_info.task_progress = sub_task_info.task_progress

		if sub_task_info.task_state == UpdateManager.TASK_STATE.COMPLETED then
JavaUtils.log("com.dayu", "Task Finish:")
			cc.Director:getInstance():getScheduler():unscheduleScriptEntry(handler)
JavaUtils.log("com.dayu", tostring(sub_task_info.task_result_code))
			if sub_task_info.task_result_code == success_condition then
JavaUtils.log("com.dayu", "Task Success")
				self:start_update(task_info)

			else

				task_info.task_progress = 0
				task_info.task_state = UpdateManager.TASK_STATE.COMPLETED
				task_info.task_result_code = sub_task_info.task_result_code

			end

		end

	end, 0.1, false)	

end

-- 更新流程的当前实现。
--
-- !! 游戏更新流程的流程图保存在tower的文档当中。
--
-- 这里使用了状态机来实现更新流程来简化代码结构。

function UpdateManager:start_update_ex()

	function enter_download_manifest()
print("download manifest start")
		local task_info = {}
		self:download_version_manifest(task_info) 

		local task_checker_handler = 0
		local task_checker = function()

			if task_info.task_state == UpdateManager.TASK_STATE.COMPLETED then

				if task_info.task_result_code == UpdateManager.RESULT_CODE.RETRIEVE_VERSION_SUCCEEDED then
print("download manifest success")
					self.__c_state_machine:do_event("yes")

				else
print("download manifest fail")
					self.__c_state_machine:do_event("no")

				end

				cc.Director:getInstance():getScheduler():unscheduleScriptEntry(task_checker_handler)

			else
print("download manifest progress")
				cc.Director:getInstance():getEventDispatcher():dispatchEvent(cc.EventCustom:new("DOWNLOAD_MANIFEST_PROGRESS"))
				print(string.format("%s%s", UpdateManager.VERSION_MANIFEST_URL, "version_manifest"))

			end

		end

		-- 启动scheduler来监测version_manifest的下载情况

		task_checker_handler = cc.Director:getInstance():getScheduler():scheduleScriptFunc(task_checker, 0.016, false)

	end

	function enter_get_local_manifest()

		-- 如果未能在线获取到version_manifest则检查本地是否有之前获取的version_mainifest

		if self:get_version_manifest() then
print("get local manifest success")
			self.__c_state_machine:do_event("yes")

		else
print("get local manifest fail")
			self.__c_state_machine:do_event("no")

		end

	end

	function enter_check_version()
print("check version")
		if self:compare_version() == UpdateManager.RESULT_CODE.CURRENT_VERSION_EXPIRED then
print("version expired")
			self.__c_state_machine:do_event("need_update")

		elseif  self:compare_version() == UpdateManager.RESULT_CODE.CURRENT_VERSION_UP_TO_DATE then
print("version update to date")
			self.__c_state_machine:do_event("up_to_date")

		elseif self:compare_version() == UpdateManager.RESULT_CODE.CURRENT_VERSION_INVALID then
print("version invalid")
			self.__c_state_machine:do_event("version_invalid")

		end

	end

	function enter_update_confirm()

		-- 若检测到当前版本属于过期版本，UpdateManager不会立即开始更新，而是会触发一个名为REQUEST_UPDATE_CONFIRM的
		-- 事件向UpdateManager的客户端请求是否执行更新，游戏实现需要接收该事件并返回对事件的处理结果（向UpdateMangaer发送事件）

print("update confirm") 
		
		-- 当UpdateManager收到UPDATE_CONFIRMED事件后，会继续更新，下载更新文件
		local confirmed_listener = cc.EventListenerCustom:create("UPDATE_CONFIRMED", function()
print("update confirmed")
			self.__c_state_machine:do_event("yes")

		end)
		cc.Director:getInstance():getEventDispatcher():addEventListenerWithSceneGraphPriority(confirmed_listener, cc.Director:getInstance():getNotificationNode())

		-- 当UpdateManager收到UPDATE_CANCELED事件后，会取消更新。
		local canceled_listener = cc.EventListenerCustom:create("UPDATE_CANCELED", function()
print("update canceled")
			self.__c_state_machine:do_event("no")

		end)
		cc.Director:getInstance():getEventDispatcher():addEventListenerWithSceneGraphPriority(canceled_listener, cc.Director:getInstance():getNotificationNode())		

		cc.Director:getInstance():getEventDispatcher():dispatchEvent(cc.EventCustom:new("REQUEST_UPDATE_CONFIRM"))

	end

	function enter_download_update()
print("download update")
		local task_info = {}
		self:start_update(task_info)

		local task_checker_handler = 0
		local task_checker = function()

			if task_info.task_state == UpdateManager.TASK_STATE.COMPLETED then

				if task_info.task_result_code == UpdateManager.RESULT_CODE.UPDATE_SUCCEEDED then
print("download update success")
					self.__c_state_machine:do_event("yes")

				else
print("download update fail")
					self.__c_state_machine:do_event("no")

				end

				cc.Director:getInstance():getScheduler():unscheduleScriptEntry(task_checker_handler)

			else
print("download update progress")
				cc.Director:getInstance():getEventDispatcher():dispatchEvent(cc.EventCustom:new("DOWNLOAD_UPDATE_PROGRESS"))

			end

		end

		task_checker_handler = cc.Director:getInstance():getScheduler():scheduleScriptFunc(task_checker, 0.016, false)		

	end

	function enter_start_game()
print("update success")
		-- 更新成功，发送UPDATE_SUCCEEDED事件
		cc.Director:getInstance():getEventDispatcher():dispatchEvent(cc.EventCustom:new("UPDATE_SUCCEEDED"))

	end

	function enter_abort_game()
print("update fail")
		-- 更新失败，发送UPDATE_FAILED事件
		cc.Director:getInstance():getEventDispatcher():dispatchEvent(cc.EventCustom:new("UPDATE_FAILED"))

	end

	self.__c_state_machine = StateMachine:create()

	-- 状态机的状态来与更新流程图中的各操作状态一致，具体参照更新流程图

	self.__c_state_machine:add_state({

		{name = "idle"}, --初始状态
		{name = "download_manifest", enter = enter_download_manifest}, --获取version_manifest
		{name = "get_local_manifest", enter = enter_get_local_manifest}, --获取本地version_manifest
		{name = "check_version", enter = enter_check_version}, --比较版本
		{name = "update_confirm", enter = enter_update_confirm}, --确认更新请求
		{name = "download_update", enter = enter_download_update}, --获取更新文件
		{name = "start_game", enter = enter_start_game}, -- 更新成功，开始游戏
		{name = "abort_game", enter = enter_abort_game}, -- 更新失败，终止游戏

	})
	self.__c_state_machine:add_transition({

		{event = "start", from = "idle", to = "download_manifest"},
		{event = "yes", from = "download_manifest", to = "check_version"},
		{event = "need_update", from = "check_version", to = "update_confirm"},
		{event = "yes", from = "update_confirm", to = "download_update"},
		{event = "yes", from = "download_update", to = "start_game"},
		{event = "no", from = "download_manifest", to = "get_local_manifest"},
		{event = "yes", from = "get_local_manifest", to = "check_version"},
		{event = "no", from = "get_local_manifest", to = "start_game"},
		{event = "version_invalid", from = "check_version", to = "abort_game"},
		{event = "up_to_date", from = "check_version", to = "start_game"},
		{event = "no", from = "update_confirm", to = "start_game"},
		{event = "no", from = "download_update", to = "abort_game"},

	})

	self.__c_state_machine:do_set("idle")

	self.__c_state_machine:do_event("start")

end

return UpdateManager

